// idpctx.cpp

#ifndef WIN32
#include <stdio.h>
#endif

#include "xmlElement.h"
#include "xmlParser.h"

#include "idpctx.h"
#include "idputil.h"
#include "idpurl.h"
#include "IcctSign.h"


void IDP_Context::Open(const STRING & Path)
{
	if(Path.empty())
		m_Path = GetDefaultConfiguration();
	else
		m_Path = Path;

	IDPTRACE(0, _T("OPEN CONTEXT"), m_Path);
		
	CreatePath(m_Path);
}


void IDP_Context::SetOption(const STRING & OptionName, const STRING & Value)
{
	if(OptionName == _T("UserName"))
		m_UserName = Value;
	else if(OptionName == _T("Password"))
		m_Password = Value;
	else if(OptionName == _T("VTAUTOURL"))
		m_VtAutoURL = Value;
	else
		IDPTHROW1(IDP_E_INVALID_OPTION);
}

void IDP_Context::XmlFromStream(STRINGSTREAM & s, xmlElement ** ppElem)
{
	xmlParser parser;
	xmlElementPtr pRoot;
	if(parser.parseText(s, pRoot.Ptr()))
		IDPTHROW1(IDP_E_INVALID_FILE);

	if(!pRoot->getChild(_T("IDP"), 0, ppElem))
		IDPTHROW1(IDP_E_INVALID_FILE);
}

void IDP_Context::XmlFromString(const STRING & s, xmlElement ** ppElem)
{
	STRINGSTREAM ea;
	ea << s.c_str();
	XmlFromStream(ea, ppElem);
}

HRESULT IDPCALL IDP_Context::XmlCreateRequest(const STRING & MsgType, xmlElement ** ppElem)
{
	xmlParser parser;
	parser.createElement(_T("IDP"), ppElem);
	(* ppElem)->setAttribute(_T("MSGTYPE"), MsgType);

	return IDP_S_OK;
}

bool IDP_Context::QueryValue(const STRING & SubPath, const STRING & SubItem, STRING & Result)
{
	if(m_Path.empty())
		IDPTHROW1(IDP_E_CONTEXT);

	STRING FilePath = m_Path + _T("/") + SubPath;

	STRINGSTREAM s;

	try
	{
		StreamFromURL(FilePath, s);
	}
	catch(IDP_Exception & e)
	{
		if(e.hr != IDP_E_RESOURCE_NOT_FOUND)
			throw e;

		return false;
	}

	if(SubItem.empty())
	{
		Result = s.str();
	}
	else
	{
		xmlElementPtr pElem;
		XmlFromStream(s, pElem.Ptr());

		if(!xmlGetAttributeEx(pElem, SubItem, Result))
			return false;

		if(Result == _T(""))
			return false;
	}

	return true;

}

bool IDP_Context::QueryValue(const STRING & SubPath, UCHAR * pData, ULONG * pcbData)
{
	if(m_Path.empty())
		IDPTHROW1(IDP_E_CONTEXT);

	STRING SubPathSl = SubPath;
#ifdef WIN32
	SlToBackSl(SubPathSl);
#endif

	STRING FilePath = m_Path + _T("/") + SubPathSl;

	FILE * f = topen(FilePath.c_str(), _T("rb"));
	if(!f)
	{
		return false;
	}

	int c = fread(pData, 1, * pcbData, f);
	if(c < 0)
	{
		fclose(f);
		IDPTHROW1(IDP_E_FILE_ERROR);
	}

	fclose(f);

	* pcbData = c;

	return true;
}

void IDP_Context::RequestToURL(const STRING & BaseUrl, STRINGSTREAM & Request, const STRING Signature, STRING & Url)
{
	STRING::size_type pos = BaseUrl.find_first_of(_T("://"), 0);
	if(pos == 0 || pos == SxString::npos)
		IDPTHROW1(IDP_E_INVALID_URL);

	pos += tcslen(_T("://"));

	STRING Protocol, Host;
	Protocol = STRING(BaseUrl, 0, pos);
	Host = STRING(BaseUrl, pos, Url.size() - pos);

	STRING Credentials;
	if(!m_UserName.empty() || !m_Password.empty())
		Credentials = m_UserName + _T(":") + m_Password + _T("@");
	
	Url = Protocol + Credentials + Host
		+ _T("?DATA=") + Encode(STRING(Request.str()));

	if(!Signature.empty())
		Url += _T("&SIGNATURE=") + Signature;
}

void IDP_Context::UpdateFile(const STRING & SubPath, UCHAR * pData, ULONG cbData)
{
	if(m_Path.empty())
		IDPTHROW1(IDP_E_CONTEXT);

	STRING SubPathSl = SubPath;
#ifdef WIN32
	SlToBackSl(SubPathSl);
#endif

	STRING FilePath = m_Path + _T("/") + SubPathSl;

	FILE * f = topen(FilePath.c_str(), _T("wb"));
	if(!f)
		IDPTHROW1(IDP_E_FILE_NOT_FOUND)

	if(fwrite(pData, cbData, 1, f) != cbData)
		IDPTHROW1(IDP_E_FILE_ERROR);

	fclose(f);

	STRING msg = _T("UPDATE: ") + SubPath;
	LogMessage(msg);
}


void IDP_Context::UpdateFile(const STRING & SubPath, const STRING & Data)
{
	if(m_Path.empty())
		IDPTHROW1(IDP_E_CONTEXT);

	STRING SubPathSl = SubPath;
#ifdef WIN32
	SlToBackSl(SubPathSl);
#endif

	STRING FilePath = m_Path + _T("/") + SubPathSl;

	FILE * f = topen(FilePath.c_str(), _T("wb"));
	if(!f)
	{
		IDPTHROW1(IDP_E_FILE_NOT_FOUND);
	}

	for(STRING::size_type i = 0; i < Data.size(); ++i)
	{
		CHAR c = (char)Data.at(i);
		if(fwrite(&c, 1, 1, f) != 1)
			IDPTHROW1(IDP_E_FILE_ERROR);
	}

	fclose(f);

	IDPTRACE(0, _T("UPDATE"), SubPath);
}

void IDP_Context::Setup()
{
	STRING url, tag;
	STRING version;
	STRING Request;
	STRINGSTREAM Response;
	STRINGSTREAM s;

	// create defaults file if not exist


	// get VtAutoUrl
	if(m_VtAutoURL.empty())
	{
		if(!QueryValue(_T("config.xml"), _T("IDP.VTAUTOURL"), url))
			IDPTHROW1(IDP_E_NO_CONFIG);
	}
	else
	{
		url = m_VtAutoURL;
	}

	// <-- SetupRequest
	xmlElementPtr pElem;
	XmlCreateRequest(_T("SetupRequest"), pElem.Ptr());

	if(QueryValue(_T("config.xml"), _T("VERSION"), version))
		pElem->setAttribute(_T("VERSION"), version);
	else
		pElem->setAttribute(_T("VERSION"), _T("0"));

	pElem->toStream(s);

	RequestToURL(url, s, _T(""), Request);

	// --> SetupResponse
	StreamFromURL(Request, Response);

	pElem = 0;
	XmlFromStream(Response, pElem.Ptr());

	CheckResponseMsgType(pElem, _T("SetupResponse"));

	// --> SetupResponse --> config.xml
	UpdateFile(_T("config.xml"), Response.str());
}

void IDP_Context::ImportKey(STRING & KeyId)
{
	STRING url;
	STRING version;
	STRING Request;
	STRINGSTREAM Response;
	STRINGSTREAM s;
	STRING KeyFile, KeyBlob;

	if(m_VtAutoURL.empty())
	{
		if(!QueryValue(_T("config.xml"), _T("IDP.VTAUTOURL"), url))
			IDPTHROW1(IDP_E_NO_CONFIG);
	}
	else
	{
		url = m_VtAutoURL;
	}

	// <-- KeyQueryRequest
	xmlElementPtr pElem;
	XmlCreateRequest(_T("KeyQueryRequest"), pElem.Ptr());

	pElem->setAttribute(_T("KEYID"), KeyId);
	pElem->toStream(s);

	RequestToURL(url, s, _T(""), Request);

	// --> KeyQueryResponse
	StreamFromURL(Request, Response);

	pElem = 0;
	XmlFromStream(Response, pElem.Ptr());

	CheckResponseMsgType(pElem, _T("KeyQueryResponse"));

	pElem->getAttribute(_T("KEYID"), KeyId);
	pElem->getAttribute(_T("KEYBLOB"), KeyBlob);

	KeyFile = _T("keys/public/") + KeyId;

	// --> KeyQueryResponse --> keys/public
	UpdateFile(KeyFile, KeyBlob);

	IDPTRACE(0, _T("IMPORT"), KeyId);
}

void IDP_Context::GenKey(STRING & KeyId)
{
	UCHAR PUB[4096];
	UCHAR PRIV[4096];
	int cbPUB, cbPRIV;

	if(ICCT_GenerateKeys(PUB, &cbPUB, PRIV, &cbPRIV, NULL, 0) != 0)
		IDPTHROW1(IDP_E_CRYPT);

	STRING PublicKey = BufferToHex(PUB, cbPUB);
	STRING PrivateKey = BufferToHex(PRIV, cbPRIV);
	STRING KeyFile;
	STRINGSTREAM s;

	KeyFile = _T("keys/public/") + KeyId;
	UpdateFile(KeyFile, PublicKey);

	KeyFile = _T("keys/private/") + KeyId;
	UpdateFile(KeyFile, PrivateKey);
}

class IDP_WCHARBuffer
{
private:
	unsigned short * p;

public:
	IDP_WCHARBuffer() { p = 0; }
	~IDP_WCHARBuffer() { delete p; }
	void Allocate(ULONG Size) { p = new unsigned short[Size]; }
	unsigned short * DataPtr() { _ASSERT(!!p); return p; }
};

void IDP_Context::VerifyData(const STRING & Data, const STRING & KeyId,
										const STRING & Signature)
{
	UCHAR pbSignature[4096]; ULONG cbSignature = sizeof(pbSignature);
	UCHAR pbKey[1024]; ULONG cbKey = sizeof(pbKey);
	STRING KeyFile, KeyBlob;
	IDP_WCHARBuffer wData;
	SX_SIZE_T i;

	KeyFile = _T("keys/public/") + KeyId;

	if(!QueryValue(KeyFile, _T(""), KeyBlob))
		IDPTHROW1(IDP_E_NOT_FOUND);

	if(HexToBuffer(KeyBlob, pbKey, &cbKey))
		IDPTHROW1(IDP_E_INVALID_KEYBLOB);

	if(HexToBuffer(Signature, pbSignature, &cbSignature))
		IDPTHROW1(IDP_E_INVALID_KEYBLOB);

	wData.Allocate(Data.size() + 1);

	for(i = 0; i < Data.size(); i++)
	{
#ifdef B_ENDIAN
		// convert to little endian representation
		wData.DataPtr()[i] = ((((unsigned short) Data.at(i)) & 0x00ff) * 0x0100)  +
			((((unsigned short) Data.at(i)) & 0xff00) / 0x0100);
#else
		wData.DataPtr()[i] = (unsigned short) Data.at(i);
#endif
	}
	wData.DataPtr()[i] = (unsigned short)0;

	if(ICCT_Verify(pbKey, cbKey, (UCHAR *) wData.DataPtr(), Data.size() * sizeof(unsigned short),
		pbSignature, cbSignature))
	{
		IDPTHROW1(IDP_E_VERIFY_FAILED);
	}

}


void IDP_Context::SignData(const STRING & Data, const STRING & KeyId, 
									  STRING & Signature)
{
	HRESULT hr;
	IDP_WCHARBuffer wData;
	UCHAR pbDigest[1024]; ULONG cbDigest = sizeof(pbDigest);
	UCHAR pbKey[1024]; ULONG cbKey = sizeof(pbKey);
	SX_SIZE_T i;
	
	STRING KeyFile, KeyBlob;
	KeyFile = _T("keys/private/") + KeyId;

	if(!QueryValue(KeyFile, _T(""), KeyBlob))
		IDPTHROW1(IDP_E_NOT_FOUND);

	wData.Allocate(Data.size() + 1);

	for(i = 0; i < Data.size(); i++)
	{
#ifdef B_ENDIAN
		// convert to little endian representation
		wData.DataPtr()[i] = ((((unsigned short)(UCHAR)Data.at(i)) & 0x00ff) * 0x0100)  +
			((((unsigned short)(UCHAR)Data.at(i)) & 0xff00) / 0x0100);
#else
		wData.DataPtr()[i] = (unsigned short) ((UCHAR) Data.at(i));
#endif
	}

	wData.DataPtr()[i] = (unsigned short)0;

	if(hr = HexToBuffer(KeyBlob, pbKey, &cbKey))
		IDPTHROW1(IDP_E_INVALID_KEYBLOB);

	if(ICCT_Signature(pbKey, cbKey, (UCHAR *) wData.DataPtr(), Data.size() * sizeof(unsigned short),
		pbDigest, &cbDigest))
	{
		IDPTHROW1(IDP_E_CRYPT);
	}

	Signature = BufferToHex(pbDigest, cbDigest);
}

void IDP_Context::LoadDefaults()
{
	STRING url;
	STRING Request;
	STRINGSTREAM Response;
	STRINGSTREAM s;

	if(m_VtAutoURL.empty())
	{
		if(!QueryValue(_T("config.xml"), _T("IDP.VTAUTOURL"), url))
			IDPTHROW1(IDP_E_NO_ENDPOINT);
	}
	else
	{
		url = m_VtAutoURL;
	}
	
	// <-- LoadDefaultsRequest
	xmlElementPtr pElem;
	XmlCreateRequest(_T("LoadDefaultsRequest"), pElem.Ptr());
	pElem->toStream(s);


	RequestToURL(url, s, _T(""), Request);

	// --> LoadDefaultsResponse
	StreamFromURL(Request, Response);

	pElem = 0;
	XmlFromStream(Response, pElem.Ptr());

	CheckResponseMsgType(pElem, _T("LoadDefaultsResponse"));

	IDPTRACE(0, _T("LOAD DEFAULTS"), _T(""));

	// --> data to default.xml
	UpdateFile(_T("default.xml"), Response.str());

}

void IDP_Context::CheckResponseMsgType(xmlElement * pElem, const STRING & CheckMsgType)
{
	STRING MsgType;
	STRING msg;
	pElem->getAttribute(_T("MSGTYPE"), MsgType);

	if(MsgType == _T("Error"))
	{
		pElem->getAttribute(_T("MESSAGE"), msg);
		IDPTHROW2(IDP_E_REQUEST_ERROR, msg);
	}
	if(!(MsgType == CheckMsgType.c_str()))
	{
		msg += _T("Response MSGTYPE: received '") + MsgType + _T("', expected '") 
			+ CheckMsgType + _T("'");
		IDPTHROW2(IDP_E_REQUEST_FAILED, msg);
	}
}

void IDP_Context::PublishKey(STRING & KeyId)
{
	STRING url;
	STRING version;
	STRING Request;
	STRINGSTREAM Response;
	STRINGSTREAM s;
	STRING KeyFile, KeyBlob, KeyExp;


	if(m_VtAutoURL.empty())
	{
		if(!QueryValue(_T("config.xml"), _T("IDP.VTAUTOURL"), url))
			IDPTHROW1(IDP_E_NO_ENDPOINT);
	}
	else
	{
		url = m_VtAutoURL;
	}

	// default valitidty: 100 days
	KeyExp = GetTimeDate(100);

	// <-- KeyExportRequest
	xmlElementPtr pElem;
	XmlCreateRequest(_T("KeyExportRequest"), pElem.Ptr());

	KeyFile = _T("keys/public/") + KeyId;
	if(!QueryValue(KeyFile, _T(""), KeyBlob))
		IDPTHROW1(IDP_E_KEY_NOT_FOUND);

	pElem->setAttribute(_T("KEYID"), KeyId);
	pElem->setAttribute(_T("KEYBLOB"), KeyBlob);
	pElem->toStream(s);

	RequestToURL(url, s, _T(""), Request);

	// --> KeyExportResponse
	StreamFromURL(Request, Response);

	pElem = 0;
	XmlFromStream(Response, pElem.Ptr());

	CheckResponseMsgType(pElem, _T("KeyExportResponse"));

	IDPTRACE(0, _T("PUBLISH"), KeyId);

	// --> KeyId to keys/current
	UpdateFile(_T("keys/current"), KeyId);
}

STRING IDP_Context::GetPath()
{
	return m_Path;
}

void IDP_Message::Create(IDP_Context * pCtx, const STRING & MsgType)
{
	m_pData = NULL;
	m_pCtx = pCtx;

	m_pCtx->XmlCreateRequest(_T("PayInit"), m_pData.Ptr());
	if(!m_pCtx->QueryValue(_T("keys/current"), _T(""), m_KeyId))
		IDPTHROW1(IDP_E_NO_KEYID);

	m_pData->setAttribute(_T("MSGTYPE"), MsgType);
	m_pData->setAttribute(_T("KEYID"), m_KeyId);

	if(MsgType == _T("PayComplete"))
	{
		m_pData->setAttribute(_T("CONFIGPATH"), m_pCtx->GetPath());
	}
}


void IDP_Message::SetAttribute(const STRING & Attribute, 
										  const STRING & Value)
{
	if(!m_pData)
		IDPTHROW1(IDP_E_UNEXPECTED);

	m_pData->setAttribute(Attribute, Value);
}

void IDP_Message::GetAttribute(const STRING & Attribute, STRING & Value)
{
	if(!m_pData)
		IDPTHROW1(IDP_E_UNEXPECTED);

	if(!m_pData->getAttribute(Attribute, Value))
		IDPTHROW1(IDP_E_NOT_FOUND);
}

void IDP_Message::GetURL(STRING & Url)
{
	if(!m_pData || !m_pCtx)
		IDPTHROW1(IDP_E_UNEXPECTED);

	STRING Signature;
	STRINGSTREAM s;
	STRING VtUrl;

	m_pData->toStream(s);

	if(!m_pCtx->QueryValue(_T("config.xml"), _T("VTURL"), VtUrl))
		IDPTHROW1(IDP_E_NO_CONFIG);

	m_pCtx->SignData(s.str(), m_KeyId, Signature);
	m_pCtx->RequestToURL(VtUrl, s, Signature, Url);
}

void IDP_Message::Capture()
{
	if(!m_pData || !m_pCtx)
		IDPTHROW1(IDP_E_UNEXPECTED);

	STRING Url(_T(""));
	STRING Signature;
	STRINGSTREAM s, r;
	STRING CaptureUrl;
	xmlElementPtr pElem;

	m_pData->toStream(s);

	if(!m_pCtx->QueryValue(_T("config.xml"), _T("CAPTUREURL"), CaptureUrl))
		IDPTHROW1(IDP_E_NO_CONFIG);

	STRING v;
	if(!m_pData->getAttribute(STRING(_T("TOKEN")), v))
		IDPTHROW1(IDP_E_UNEXPECTED);

	m_pCtx->SignData(s.str(), m_KeyId, Signature);
	m_pCtx->RequestToURL(CaptureUrl, s, Signature, Url);

	m_pCtx->StreamFromURL(Url, r);
	m_pCtx->XmlFromStream(r, pElem.Ptr());

	m_pCtx->CheckResponseMsgType(pElem, _T("PayConfirm"));

	/*STRING MsgType;
	pElem->getAttribute(_T("MSGTYPE"), MsgType);
	if(MsgType != _T("PayConfirm"))
	{
		IDPTHROW1(IDP_E_REQUEST_FAILED);
	}*/
}


void IDP_Context::Reconfigure()
{
	if(m_Path.empty())
		IDPTHROW1(IDP_E_CONTEXT);

	// create directories
	CreatePath(m_Path + _T("/keys"));
	CreatePath(m_Path + _T("/keys/private"));
	CreatePath(m_Path + _T("/keys/public"));

	// run reconfiguration
	IDPTRACE(0, _T("START RECONFIGURE"), _T(""));

	STRING tag;
	bool ResetDefaults = false;

	if(!QueryValue(_T("default.xml"), _T("IDP.ISVALID"), tag))
		ResetDefaults = true;
	else if(tag != _T("yes"))
		ResetDefaults = true;

	// update config.xml
	Setup();

	if(ResetDefaults)
	{
		LoadDefaults();
	}

	STRING VtKeyId;
	if(!QueryValue(_T("config.xml"), _T("VTKEYID"), VtKeyId))
		IDPTHROW1(IDP_E_NO_CONFIG);

	ImportKey(VtKeyId);

	// create keys
	STRING Version, CustomerId;
	QueryValue(_T("config.xml"), _T("CUSTOMERID"), CustomerId);
	QueryValue(_T("config.xml"), _T("VERSION"), Version);

	STRING KeyId = _T("0-") + CustomerId + _T("-") + CreateUuid();

	GenKey(KeyId);

	// publish key
	PublishKey(KeyId);

	IDPTRACE(0, _T("RECONFIGURE OK"), _T(""));
}

void IDP_Context::LogMessage(const STRING & msg)
{
	return;

#ifdef _DEBUG
	tprintf(_T("%s\n"), msg.c_str());
#endif

#ifdef WIN32
	FILE * f = fopen("c:\\temp\\idpapp.out", "a");
#else
	FILE * f = fopen("/tmp/idpapp.out", "a");
#endif
	if(f)
	{
#ifdef UNICODE
		fprintf(f, "%S\n", msg.c_str());
#else
		fprintf(f, "%s\n", msg.c_str());
#endif
		fclose(f);
	}
}

void IDP_Message::Open(IDP_Context * pCtx, const STRING & Data, const STRING & Signature)
{
	m_pData = NULL;
	m_pCtx = pCtx;
	
	m_pCtx->XmlFromString(Data, m_pData.Ptr());

	if(Signature.empty())
	{
		IDPTHROW1(E_NO_SIGNATURE);
	}

	STRING KeyId;
	if(!m_pData->getAttribute(_T("KEYID"), KeyId))
		IDPTHROW1(IDP_E_NO_KEYID);

	m_pCtx->VerifyData(Data, KeyId, Signature);

	STRING MsgType;
	if(m_pData->getAttribute(_T("MSGTYPE"), MsgType))
	{
		if(MsgType == _T("PayConfirm"))
		{
			m_pData->setAttribute(_T("CONFIGPATH"), m_pCtx->GetPath());
		}
	}
}

void IDP_Context::SetValue(const STRING & SubPath, const STRING & SubItem, const STRING & Value,
						   bool CreateFileIfNotExist)
{
	STRING s; xmlElementPtr pElem; xmlParser parser;
	QueryValue(SubPath, _T(""), s);

	// create file only if option is set
	if(CreateFileIfNotExist)
	{
		try
		{
			XmlFromString(s, pElem.Ptr());
		}
		catch(IDP_Exception & e)
		{
			if(e.hr != IDP_E_INVALID_FILE)
				throw e;

			parser.createElement(_T("IDP"), pElem.Ptr());
		}
	}
	else
	{
		XmlFromString(s, pElem.Ptr());
	}

	xmlSetAttributeEx(pElem, SubItem, Value);
	STRINGSTREAM t; pElem->toStream(t);
	UpdateFile(SubPath, t.str()); 
}

void IDP_Message::Request(IDP_Message & msg)
{
	if(!m_pData || !m_pCtx)
		IDPTHROW1(IDP_E_UNEXPECTED);

	STRING MsgType;
	m_pData->getAttribute(_T("MSGTYPE"), MsgType);
	m_pData->setAttribute(_T("MSGTYPE"), MsgType + _T("Request"));

	// expected response
	STRING Url(_T(""));
	STRING Signature;
	STRINGSTREAM s, r;
	STRING CaptureUrl;
	xmlElementPtr pElem;

	m_pData->toStream(s);

	if(!m_pCtx->QueryValue(_T("config.xml"), _T("CAPTUREURL"), CaptureUrl))
		IDPTHROW1(IDP_E_NO_CONFIG);

	m_pCtx->SignData(s.str(), m_KeyId, Signature);
	m_pCtx->RequestToURL(CaptureUrl, s, Signature, Url);

	m_pCtx->StreamFromURL(Url, r);
	m_pCtx->XmlFromStream(r, pElem.Ptr());

	m_pCtx->CheckResponseMsgType(pElem, MsgType + _T("Response"));

	msg.m_pData = pElem;
}
