// idperr.cpp

#include <stdio.h>

#include "idpctx.h"
#include "idperr.h"
#include "idputil.h"

#ifdef WIN32
#define IDPEXPORT __declspec(dllexport)
#else
#define IDPEXPORT
#endif

struct IDP_ErrorEntry
{
	HRESULT hr;
	LPCTSTR pszMessage;
};

IDP_ErrorEntry ErrorTable[] =
{
	{ IDP_E_CONTEXT,			_T("no context present") },
	{ IDP_E_FILE_NOT_FOUND,		_T("file not found") },
	{ IDP_E_FILE_ERROR,			_T("file I/O error") },
	{ IDP_E_INVALID_FILE,		_T("invalid file format") },
	{ IDP_E_BUFFER_SIZE,		_T("the buffer size is too small") },
	{ IDP_E_INVALID_URL,		_T("invalid URL") },
	{ IDP_E_INVALID_OPTION,		_T("invalid option") },
	{ IDP_E_REQUEST_FAILED,		_T("request failed") },
	{ IDP_E_CRYPT,				_T("cryptographic error") },
	{ IDP_E_NO_ENDPOINT,		_T("no endpoint") },
	{ IDP_E_INVALID_KEYBLOB,	_T("invalid key blob") },
	{ IDP_E_FAIL,				_T("command failed") },
	{ IDP_E_NOT_FOUND,			_T("requested item was not found") },
	{ IDP_E_NO_KEYID,			_T("no key identifier specified") },
	{ IDP_E_NO_CONFIG,			_T("no configuration found") },
	{ IDP_E_KEY_NOT_FOUND,		_T("the requested key was not found") },
	{ IDP_E_UNEXPECTED,			_T("unexpected error") },
	{ IDP_E_NO_DEFAULT_PATH,	_T("no default path was supplied") },
	{ IDP_E_REQUEST_ERROR,		_T("the request failed") },
	{ IDP_E_VERIFY_FAILED,		_T("invalid signature") },
	{ E_NO_SIGNATURE,			_T("missing signature") },

	// URLMON.DLL results
	{ 0x80070005,				_T("access to URL resource denied") },
	{ 0x800c0005,				_T("URL resource not found") },
	{ 0x800c0006,				_T("URL resource not found") },
	{ 0x800c000c,				_T("invalid URL request") },

	// socket results
	{ IDP_SOCKET_BASE + 61,		_T("connection refused") },
	{ IDP_SOCKET_BASE + 60,		_T("connection timeout") },
	{ IDP_SOCKET_BASE + 54,		_T("connection reset") },
	{ IDP_SOCKET_BASE + 53,		_T("connection aborted") },

	{ IDP_SOCKET_BASE + 51,		_T("network not reachable") },

	// HTTP
	{ IDP_HTTP_DNS_ERROR,		_T("DNS lookup failed") },
	{ IDP_HTTP_BASE + 401,		_T("access to URL resource denied") },
	{ IDP_HTTP_BASE + 404,		_T("URL resource not found") },
	{ IDP_HTTP_BASE + 503,		_T("server temporarily unavailable") },

	{ 0 }

};

bool IDP_ErrorLookup::GetErrorStringFromResult(HRESULT hr, STRING & txt)
{
	IDP_ErrorEntry * pTable = ErrorTable;
	while(pTable->hr != 0)
	{
		if(pTable->hr == hr)
		{
			txt = pTable->pszMessage;
			return true;
		}

		pTable++;
	}

	return false;

}

bool IDP_ErrorLookup::UseExtendedErrorInfo()
{
	LPCSTR p = getenv("IDPDEBUG");
	if(p)
		return true;
	else
		return false;
}

extern "C" IDPEXPORT long IDP_GetErrorStringFromResult(long hr, long *pcbSize, char *pTxt)
{
	STRING sTxt;
	if(IDP_ErrorLookup::GetErrorStringFromResult(hr, sTxt))
	{
		char *buf = new char[sTxt.size()*2+2];	// MBCS
		HRESULT hr = 0;
#ifdef UNICODE
		sprintf(buf, "%S", sTxt.c_str());
#else
		sprintf(buf, "%s", sTxt.c_str());
#endif
		long size = strlen(buf)+1;
		if(pcbSize)
		{
			if(*pcbSize > size)
			{
				if(pTxt) strcpy(pTxt, buf);
			}
			else if(*pcbSize)
			{
				if(pTxt)
				{
					strncpy(pTxt, buf, (*pcbSize)-1);
					pTxt[(*pcbSize)-1] = '\0';
				}
			}
			*pcbSize = size;
			delete buf;
			return true;
		}
		delete buf;
	}
	return false;
}


IDP_Exception::IDP_Exception(HRESULT newHr, const STRING & newErrorString, 
							 LPCSTR newSourceFile, long newLineNumber)
{
	STRINGSTREAM msg;
	hr = newHr;
	TCHAR tHr[9];
	tsprintf(tHr, _T("%x"), newHr);
	STRING txt;

	IDP_ErrorLookup::GetErrorStringFromResult(newHr, txt);

#ifdef UNICODE
	WCHAR tmp[1024];
	mbstowcs(tmp, newSourceFile, 1024);
#else
	LPCSTR tmp = newSourceFile;
#endif
	if(IDP_ErrorLookup::UseExtendedErrorInfo())
	{
		msg << tmp << _T("(") << newLineNumber << _T(") : error ")
			<< tHr << _T(" : ");
	}
	
	if(txt.size())
	{
		msg << txt;
	}
	else
	{
		msg << _T("ErrorCode = 0x") << tHr;
	}
	
	if(newErrorString.size() > 0)
	{
		msg << _T(" (") << newErrorString.c_str() << _T(")");
	}
	ErrorString = msg.str();
	SourceFile = tmp;
	LineNumber = newLineNumber;

	IDP_TraceLog::Write(newSourceFile, newLineNumber, newHr, _T("EXCEPTION"), ErrorString);
}

void IDP_TraceLog::Write(LPCSTR ModuleArg, long LineNumber, HRESULT hr, const STRING & Action, 
	const STRING & Text)
{
#ifndef _DEBUG
	return;
#endif

	TCHAR *Module = new TCHAR[strlen(ModuleArg) + 1];
#ifdef UNICODE
	swprintf(Module, L"%S", ModuleArg);
#else
	sprintf(Module, "%s", ModuleArg);
#endif
	tprintf(_T("[%s] 0x%08x : %s\n    %s(%ld)\n"), GetTimeDate(0).c_str(), hr, Action.c_str(),
		Module, LineNumber);

	delete Module;

	if(!Text.empty())
		tprintf(_T("%s\n"), Text.c_str());

	tprintf(_T("\n"));
}
