/********************** saferpay.c: Saferpay command line utility ************************/

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <malloc.h>
#include "idpapi.h"
#include "idperrc.h"



#ifndef NULL
#define NULL ((void*)0)
#endif

#define MAX_ATTRIBUTES 100

#define LEAVE1 { goto FINALLY; }
#define LEAVE2(res) { hr = res; LEAVE1; }
#define LEAVE3(res, msg) { fprintf(stderr, "%s", msg); LEAVE2(res); }
#define CHECK_SUB(func) { if((hr = func) != IDP_S_OK) LEAVE1; }
#define CHECK_SUB2(func) { if((hr = func) != IDP_S_OK) { handleError(hr); LEAVE1; } }

#define INIT_ATTR(a) { int c; for(c = 0; c < sizeof(a)/sizeof(a[0]) - 1; ++c) { a[c].shortName = (const char*)1; } a[c].shortName = NULL; }


typedef struct _PARAMETER
{
	const char *name;
	const char *shortName;
	const char *value;
} PARAMETER;


/* prototypes */
static void handleError(HRESULT hr);
static HRESULT printObject(IDP_HANDLE hMO);
static HRESULT saveObject(IDP_HANDLE hMO, const char *pszFileName);
static HRESULT loadObject(IDP_HANDLE hMO, const char *pszFileName);
static HRESULT handleArgs(int argc, char **argv, PARAMETER **needed,
				   PARAMETER **opt, PARAMETER *attr);
static HRESULT configure(char **argv, int argc);
static HRESULT setDefaultAttributes(char **argv, int argc);
static HRESULT createPayInit(char **argv, int argc);
static HRESULT verifyPayConfirm(char **argv, int argc);
static HRESULT capture(char **argv, int argc);
static HRESULT createRequest(char **argv, int argc);
static void displayUsage();


static PARAMETER g_path = { "--path", "-p", NULL };
static PARAMETER g_serverURL = { "--serverurl", "-r", NULL };
static PARAMETER g_userName = { "--username", "-u", NULL };
static PARAMETER g_password = { "--password", "-w", NULL };
static PARAMETER g_attribute = { "--attribute", "-a", NULL };
static PARAMETER g_signature = { "--signature", "-s", NULL };
static PARAMETER g_data = { "--data", "-d", NULL };
static PARAMETER g_token = { "--token", "-t", NULL };
static PARAMETER g_id = { "--identifier", "-i", NULL };
static PARAMETER g_messageType = { "--messagetype", "-m", NULL };
static PARAMETER g_fileName = { "--filename", "-f", NULL };
static PARAMETER g_last = { NULL, NULL, NULL };


void handleError(HRESULT hr)
{
	char szMsg[1024];
	long len = sizeof(szMsg);

	if(IDP_GetErrorStringFromResult(hr, &len, szMsg))
		fprintf(stderr, "[saferpay] ERROR: 0x%08X (%s)\n", hr, szMsg);
	else
		fprintf(stderr, "ERROR: %08x\n", hr);
}

static HRESULT printObject(IDP_HANDLE hMO)
{
	HRESULT hr = IDP_S_OK;
	char *pData = NULL;	long len = 0;

	MessageObject_ToString(hMO, &len, NULL);
	if(!(pData = (char*)malloc(len))) { LEAVE2(IDP_E_OUT_OF_MEMORY) };

	CHECK_SUB(MessageObject_ToString(hMO, &len, pData));

	printf("%s", pData);

FINALLY:
	if(pData) free(pData);

	return hr;
}

static HRESULT saveObject(IDP_HANDLE hMO, const char *pszFileName)
{
	FILE *pFile = NULL;
	HRESULT hr = IDP_S_OK;
	char *pData = NULL;	long len = 0;

	MessageObject_ToString(hMO, &len, NULL);
	if(!(pData = (char*)malloc(len))) { LEAVE2(IDP_E_OUT_OF_MEMORY) };

	CHECK_SUB(MessageObject_ToString(hMO, &len, pData));

	if(!(pFile = fopen(pszFileName, "w"))) { LEAVE2(errno) };
	if(fwrite(pData, len-1, 1, pFile) != 1) { LEAVE2(errno) };


FINALLY:
	if(pFile) fclose(pFile);
	if(pData) free(pData);

	return hr;
}

static HRESULT loadObject(IDP_HANDLE hMO, const char *pszFileName)
{
	FILE *pFile = NULL;
	HRESULT hr = IDP_S_OK;
	char *pData = NULL;	long len = 0;

	if(!(pFile = fopen(pszFileName, "r"))) { LEAVE2(IDP_E_FILE_NOT_FOUND) };
	fseek(pFile, 0, SEEK_END);
	len = ftell(pFile);
	fseek(pFile, 0, SEEK_SET);
	if(!(pData = (char*)malloc(len+1))) { LEAVE2(IDP_E_OUT_OF_MEMORY) };
	if(fread(pData, len, 1, pFile) != 1) { LEAVE2(errno) };
	pData[len] = 0;

	CHECK_SUB(MessageObject_FromString(hMO, pData));

FINALLY:
	if(pFile) fclose(pFile);
	if(pData) free(pData);

	return hr;
}


/* attr->shortName == NULL indicates the end of the attr array */
static HRESULT handleArgs(int argc, char **argv, PARAMETER **needed,
				   PARAMETER **opt, PARAMETER *attr)
{
	HRESULT hr = IDP_S_OK;
	PARAMETER **cur, *a;
	char szMsg[1024];
	int i = 1, j;
    PARAMETER **param[2];
    param[0] = needed;
    param[1] = opt;

	/* initialize the arrays */
	for(j = 0; j < 2; j++)
	{
		if((cur = param[j]))
		{
			while((*cur)->name)
			{
				(*cur)->value = NULL;
				cur++;
			}
		}
	}
	if((a = attr))
	{
		while(a->shortName)
		{
			a->value = NULL;
			a->name = NULL;
			a++;
		}
	}

	/* fetch the command line arguments */
	while(++i < argc)
	{
		if(attr && attr->shortName)
		{
			if((!strcmp(argv[i], "--attribute")) || (!strcmp(argv[i], "-a")))
			{
				if(++i >= argc)
				{
					LEAVE3(IDP_E_PARAMETER, "invalid parameter format\n");
				}
				attr->name = argv[i];

				if(++i >= argc)
				{
					LEAVE3(IDP_E_PARAMETER, "invalid parameter format\n");
				}
				attr->value = argv[i];

				attr++;
				goto NEXT;
			}
		}

		{

			for(j = 0; j < 2; j++)
			{
				if((cur = param[j]))
				{
					while((*cur)->name)
					{
						if((!strcmp(argv[i], (*cur)->shortName))
							|| (!strcmp(argv[i], (*cur)->name)))
						{
							if(++i >= argc)
							{
								LEAVE3(IDP_E_PARAMETER, "invalid parameter format\n");
							}
							
							(*cur)->value = argv[i];
							goto NEXT;
						}

						*cur++;
					}
				}
			}
		}

		sprintf(szMsg, "unknown option: %s\n", argv[i]);
		LEAVE3(IDP_E_PARAMETER, szMsg);

NEXT:
		;
	}

	/* check whether all required parameters have been found */
	if((cur = needed))
	{
		while((*cur)->name)
		{
			if(!(*cur)->value)
			{
				sprintf(szMsg, "ERROR: required parameter %s is missing\n", (*cur)->name);
				LEAVE3(IDP_E_PARAMETER, szMsg);
			}

			*cur++;
		}
	}

FINALLY:
	return hr;
}


static HRESULT configure(char **argv, int argc)
{
	HRESULT hr = IDP_S_OK;
	IDP_HANDLE hCS = NULL;

	PARAMETER *needed[] = { &g_path, &g_serverURL, &g_userName, &g_password, &g_last };
	PARAMETER attr[MAX_ATTRIBUTES], *a;
	INIT_ATTR(attr);

	CHECK_SUB(handleArgs(argc, argv, needed, NULL, attr));

	/* IDP ConfigurationSetup handle */
	if(!(hCS = CreateConfigurationSetup()))
		LEAVE2(IDP_E_OUT_OF_MEMORY);

	/* configuration setup */
	CHECK_SUB2(ConfigurationSetup_SetConfiguration(hCS, g_path.value));

	/* (re-)configure */
	CHECK_SUB2(ConfigurationSetup_Reconfigure(hCS, g_serverURL.value,
		g_userName.value, g_password.value));
	
	/* set default attributes */
	a = attr;
	while(a->name)
	{
		CHECK_SUB2(ConfigurationSetup_SetAttribute(hCS, "default.xml", a->name, a->value));
		a++;
	}
	
FINALLY:
	if(hCS) DestroyConfigurationSetup(hCS);

	return hr;
}

static HRESULT setDefaultAttributes(char **argv, int argc)
{
	HRESULT hr = IDP_S_OK;
	IDP_HANDLE hCS = NULL;

	PARAMETER *needed[] = { &g_path, &g_last };
	PARAMETER attr[100], *a;
	INIT_ATTR(attr);

	CHECK_SUB(handleArgs(argc, argv, needed, NULL, attr));

	/* IDP ConfigurationSetup handle */
	if(!(hCS = CreateConfigurationSetup()))
		LEAVE2(IDP_E_OUT_OF_MEMORY);
	
	/* configuration setup */
	CHECK_SUB2(ConfigurationSetup_SetConfiguration(hCS, g_path.value));

	a = attr;
	while(a->name)
	{
		CHECK_SUB2(ConfigurationSetup_SetAttribute(hCS, "default.xml", a->name, a->value));
		a++;
	}
	
FINALLY:
	if(hCS) DestroyConfigurationSetup(hCS);
	
	return hr;
}

static HRESULT createPayInit(char **argv, int argc)
{
	HRESULT hr = IDP_S_OK;
	IDP_HANDLE hMF = NULL, hMO = NULL;
	char *pURL = NULL; long cbURL;

	PARAMETER *needed[] = { &g_path, &g_last };
	PARAMETER attr[100], *a;
	INIT_ATTR(attr);

	CHECK_SUB(handleArgs(argc, argv, needed, NULL, attr));
		
	/* IDP MessageFactory handle */
	if(!(hMF = CreateMessageFactory()))
		LEAVE2(IDP_E_OUT_OF_MEMORY);

	/* open message factory */
	CHECK_SUB2(MessageFactory_Open(hMF, g_path.value));
	
	/* create a PayInit message */
	CHECK_SUB2(MessageFactory_CreatePayInit(hMF, &hMO));
		
	/* message attributes */
	a = attr;
	while(a->name)
	{
		CHECK_SUB2(MessageObject_SetAttribute(hMO, a->name, a->value));
		a++;
	}
		
	/* get the PayInit message */
	cbURL = 1;
	MessageObject_GetURL(hMO, &cbURL, pURL);
	if(!(pURL = (char*)malloc(cbURL)))
		LEAVE2(IDP_E_OUT_OF_MEMORY);

	CHECK_SUB2(MessageObject_GetURL(hMO, &cbURL, pURL));

	printf("%s", pURL);

FINALLY:
	if(hMF) DestroyMessageFactory(hMF);
	if(hMO) DestroyMessageObject(hMO);
	if(pURL) free(pURL);

	return hr;
}

static HRESULT verifyPayConfirm(char **argv, int argc)
{
	HRESULT hr = IDP_S_OK;
	IDP_HANDLE hMF = NULL, hMO = NULL;

	PARAMETER *needed[] = { &g_path, &g_data, &g_signature, &g_last };
	PARAMETER *opt[] = { &g_fileName, &g_last };

	CHECK_SUB(handleArgs(argc, argv, needed, opt, NULL));

	/* IDP MessageFactory handle */
	if(!(hMF = CreateMessageFactory()))
		LEAVE2(IDP_E_OUT_OF_MEMORY);
	
	/* open message factory */
	CHECK_SUB2(MessageFactory_Open(hMF, g_path.value));

	/* verify PayConfirm */
	CHECK_SUB2(MessageFactory_VerifyPayConfirm(hMF, g_data.value, g_signature.value, &hMO));
	
	if(g_fileName.value)
	{
		CHECK_SUB(saveObject(hMO, g_fileName.value));
	}
	else
	{
		CHECK_SUB(printObject(hMO));
	}

FINALLY:
	if(hMF) DestroyMessageFactory(hMF);
	if(hMO) DestroyMessageObject(hMO);

	return hr;
}

static HRESULT capture(char **argv, int argc)
{
	HRESULT hr = IDP_S_OK;
	IDP_HANDLE hMF = NULL, hMO = NULL;

	PARAMETER *needed[] = { &g_path, &g_last };
	PARAMETER *opt[] = { &g_fileName, &g_id, &g_token, &g_last };
	PARAMETER attr[100], *a;
	INIT_ATTR(attr);

	CHECK_SUB(handleArgs(argc, argv, needed, opt, attr));

	/* check initializiation */
	if(g_id.value == NULL && g_fileName.value == NULL)
	{
		LEAVE3(IDP_E_PARAMETER, "ERROR: missing argument --identifier\n");
	}
	if(g_token.value == NULL && g_fileName.value == NULL)
	{
		LEAVE3(IDP_E_PARAMETER, "ERROR: missing argument --token\n");
	}
	if((g_fileName.value) && (g_token.value || g_id.value))
	{
		LEAVE3(IDP_E_PARAMETER,
			"ERROR: specify either --filename or --identifier and --token\n");
	}

	/* IDP MessageFactory handle */
	if(!(hMF = CreateMessageFactory()))
		LEAVE2(IDP_E_OUT_OF_MEMORY);

	/* open message factory */
	CHECK_SUB2(MessageFactory_Open(hMF, g_path.value));

	if(g_fileName.value != NULL)
	{
		char Token[1024], Id[1024];
		long cbToken = sizeof(Token);
		long cbId = sizeof(Id);

		/* IDP MessageObject handle */
		if(!(hMO = CreateMessageObject()))
			LEAVE2(IDP_E_OUT_OF_MEMORY);

		CHECK_SUB2(MessageObject_Open(hMO, g_path.value));
		CHECK_SUB2(loadObject(hMO, g_fileName.value));

		/* get Token */
		memset(Token, 0, cbToken);
		CHECK_SUB2(MessageObject_GetAttribute(hMO, "TOKEN", &cbToken, Token));

		/* get Id */
		memset(Id, 0, cbId);
		CHECK_SUB2(MessageObject_GetAttribute(hMO, "ID", &cbId, Id));

		/* old message object no more needed */
		DestroyMessageObject(hMO); hMO = NULL;

		/* create pay complete (Id and Token from a file) */
		CHECK_SUB2(MessageFactory_CreatePayComplete(hMF, Id, Token, &hMO));
	}
	else
	{
		/* create the PayComplete message */
		CHECK_SUB2(MessageFactory_CreatePayComplete(hMF, g_id.value, g_token.value, &hMO));
	}

	/* message attributes */
	a = attr;
	while(a->name)
	{
		CHECK_SUB2(MessageObject_SetAttribute(hMO, a->name, a->value));
		a++;
	}

	/* financial transfer */
	CHECK_SUB2(MessageObject_Capture(hMO));

FINALLY:
	if(hMF) DestroyMessageFactory(hMF);
	if(hMO) DestroyMessageObject(hMO);

	return hr;
}

static HRESULT createRequest(char **argv, int argc)
{
	HRESULT hr = IDP_S_OK;
	IDP_HANDLE hMF = NULL, hMO = NULL, hMR = NULL;

	PARAMETER *needed[] = { &g_path, &g_messageType, &g_last };
	PARAMETER *opt[] = { &g_fileName, &g_last };
	PARAMETER attr[100], *a;
	INIT_ATTR(attr);

	CHECK_SUB(handleArgs(argc, argv, needed, opt, attr));

	/* IDP MessageFactory handle */
	if(!(hMF = CreateMessageFactory()))
		LEAVE2(IDP_E_OUT_OF_MEMORY);

	/* open message factory */
	CHECK_SUB2(MessageFactory_Open(hMF, g_path.value));

	/* create a new request */
	CHECK_SUB2(MessageFactory_CreateRequest(hMF, g_messageType.value, &hMO));

	/* message attributes */
	a = attr;
	while(a->name)
	{
		CHECK_SUB2(MessageObject_SetAttribute(hMO, a->name, a->value));
		a++;
	}

    CHECK_SUB2(MessageObject_Execute(hMO, &hMR));

    if(g_fileName.value)
    {
		CHECK_SUB2(saveObject(hMR, g_fileName.value));
    }
	else
	{
		CHECK_SUB2(printObject(hMR));
	}

FINALLY:
    if(hMF) DestroyMessageFactory(hMF);
    if(hMO) DestroyMessageObject(hMO);
    if(hMR) DestroyMessageObject(hMR);

    return hr;

}

static void displayUsage()
{
	printf(
		"\nSAFERPAY COMMAND LINE UTILITY"
		"\n"
		"\n       syntax: saferpay option [arguments]"
		"\n"
		"\nThe saferpay command line utility allows access to the Saferpay Application"
		"\nComponent. The SAC exposes a set of interfaces for transaction processing"
		"\nand administration."
		"\n"
		"\n"
		"\n      Configuration options:"
		"\n               -conf               Starts the configuration process"
		"\n               -attr               Sets default attributes"
		"\n"
		"\n      Transaction options:"
		"\n               -payinit            Creates a digital offer"
		"\n               -payconfirm         Verifies the saferpay reservation"
		"\n               -capt               Financial transfer"
		"\n"
		"\n      Additional options:"
		"\n               -exec               General message exchange"
		"\n               -h, --help          Shows this help"
		"\n"
		"\n"
		"\n      Arguments:"
		"\n               -p, --path          path to your saferpay component"
		"\n               -r, --serverurl    URL to the saferpay server"
		"\n               -u, --username      unique identification of each user"
		"\n               -w, --password"
		"\n               -a, --attribute    saferpay attributes"
		"\n               -s, --signature"
		"\n               -d, --data          XML encoded string"
		"\n               -t, --token         aditional signed data"
		"\n               -i, --identifier    unique transaction identifier"
		"\n               -m, --messagetype  saferpay message type"
		"\n               -f, --filename"
		"\n"
		"\n"
		"\n"
		"\nUSAGE:"
		"\n-----"
		"\n"
		"\n   Configuration:"
		"\n   -------------"
		"\n         Defines the path for the configuration store, creates"
		"\n         or updates the saferpay client module in the specified"
		"\n         path and sets the default attributes."
		"\n"
		"\n"
		"\n      syntax:"
		"\n         saferpay -conf --path --serverurl --username --password \\"
		"\n                  { --attribute AttributeName DefaultValue }"
		"\n"
		"\n      e.g. saferpay -conf -p /opt/saferpay \\"
		"\n                    -r https://www.saferpay.com/user/setup.asp \\"
		"\n                    -u e99867001 -w XAjc3Kna"
		"\n"
		"\n           ServerURL: <https://www.saferpay.com/user/setup.asp>"
		"\n                      as default."
		"\n"
		"\n           test password and username..."
		"\n"
		"\n"
		"\n      Set default attributes seperately:"
		"\n"
		"\n      syntax:"
		"\n         saferpay -attr --path { --attribute AttributeName DefaultValue }"
		"\n"
		"\n      e.g. saferpay -attr -p /opt/saferpay -a DELEVERY no ALLOWCOLLECT no"
		"\n"
		"\n"
		"\n"
		"\n   Digital offer (PayInit message creation):"
		"\n   -------------"
		"\n         The digital offer is the starting point of the transaction."
		"\n         It is a piece of data digitally signed by the issuing merchant"
		"\n         and packed into an URL called \"PayInit message\". Needs"
		"\n         all the mandatory PayInit attributes, which are not set in"
		"\n         the default attributes (/{SAFERPAY_DIR}/default.xml)."
		"\n"
		"\n"
		"\n      syntax:"
		"\n         saferpay -payinit --path { --attribute AttributeName Value }"
		"\n"
		"\n"
		"\n      e.g. saferpay -payinit -p /opt/saferpay -a AMOUNT 1560 CURRENCY EUR \\"
		"\n                    DESCRIPTION \"Test Purchase with saferpay\" \\"
		"\n                    SUCCESSLINK <success link> BACKLINK <back link>"
		"\n                    FAILLINK <fail link> ACCOUNTID 99867-94913159"
		"\n"
		"\n           SUCCESSLINK: A URL identifing the web page to dispay after the"
		"\n                        reservation has been succesfully completed. The"
		"\n                        comfirmation message will be appended to this URL."
		"\n"
		"\n           FAILLINK:    This parameter indicates the page to display after"
		"\n                        a reservation attempt has failed."
		"\n"
		"\n           BACKLINK:    This page is displayed in the client's browser if"
		"\n                        the transaction is aborted."
		"\n"
		"\n"
		"\n"
		"\n   Reservation (PayConfirm message verification):"
		"\n   -----------"
		"\n         Checks the PayInit message, which was sent by the Saferpay"
		"\n         host. Parameters must not be URL-encoded."
		"\n"
		"\n"
		"\n      syntax:"
		"\n         saferpay -payconfirm --path --data --signature { --filename }"
		"\n"
		"\n      e.g. saferpay -payconfirm -p /opt/saferpay/ -d < datastring > \\"
		"\n                     -s < signature >"
		"\n"
		"\n			If the message have been verified successfully, the XML-encoded"
		"\n         data string could be stored in a file for a later use (eg. capturing)."
		"\n         In this case specify the path with the '--filename' argument."
		"\n"
		"\n"
		"\n"
		"\n   Finacial transfer (Capturing):"
		"\n   -----------------"
		"\n         Creates the PayComplete message, sends it to the Saferpay"
		"\n         host and checks the response. Id and Token can be taken"
		"\n         from the PayConfirm Message or from a file if a valid argument"
		"\n         '--filename' is specified (Id+Token=Transaction Handle)."
		"\n"
		"\n"
		"\n      syntax:"
		"\n         saferpay -capt --path --identifier --token"
		"\n                  { --attribute AttributeName Value }"
		"\n"
		"\n         saferpay -capt --path --filename { --attribute AttributeName Value }"
		"\n"
		"\n      e.g. saferpay -capt -p /opt/saferpay/ -f ./confirmed.log"
		"\n"
		"\n"
		"\n"
		"\n   General message exchange: Creates a Saferpay request,"
		"\n   ------------------------  adds additional attributes,"
		"\n                             sends it to the host and gets the response."
		"\n"
		"\n       saferpay -exec -p Path -m MsgType [-f FileName] {-a AttributeName Value }"
                "\n       e.g. saferpay -exec -p /opt/saferpay -m Authorization -f ./exec.log -a ATTRIBUTES"
		"\n"
		"\n"
		"\n"
		"\nRETURN VALUES:"
		"\n-------------"
		"\n   saferpay returns 0 on success, 1 on wrong usage, else the error code"
		"\n   returned by the Saferpay API.\n");
}

int main(int argc, char *argv[])
{
	if(argc < 2)
	{
		fprintf(stderr, "Error: not enough parameters specified;\n");
		fprintf(stderr, "type saferpay -h to view usage\n");
		return 1;
	}
	if((!strcmp(argv[1], "-h")) || (!strcmp(argv[1], "--help")))
	{
		displayUsage();
		return 1;
	}

	/* saferpay options */
	if(strcmp(argv[1], "-conf") == 0)
	{
		/* saferpay configuration */
		return configure(argv, argc);
	}
	else if(strcmp(argv[1], "-attr") == 0)
	{
		/* set default attributes (default.xml) */
		return setDefaultAttributes(argv, argc);
	}	
	else if(strcmp(argv[1], "-payinit") == 0)
	{
		/* create a PayInit message */
		return createPayInit(argv, argc);
	}
	else if(strcmp(argv[1], "-payconfirm") == 0)
	{	
		/* verify the PayConfirm message */
		return verifyPayConfirm(argv, argc);
	}
	else if(strcmp(argv[1], "-capt") == 0)
	{
		/* financial transfer */
		return capture(argv, argc);
	}
	else if(strcmp(argv[1], "-exec") == 0)
	{
		/* universal exec-function */
		return createRequest(argv, argc);
	}
	else
	{
		/* unknown option */
		fprintf(stderr, "unknown option %s\n", argv[1]);
		return 1;
	}
}

