// sxmap.cpp: multimap implementation

#include "sxexcept.h"
#include "sxfcerr.h"
#include "sxstack.h"
#include "sxmap.h"


long ROUND_LOG(long n)
{
	long res = -1;

	do
	{
		n /= 2;
		res++;
	} while(n);

	return res;
}

/////////////// implementation of class SxTreeNode


// construction/destruction

SxTreeNode::SxTreeNode()
{
	m_pLeft = 0;
	m_pRight = 0;
	m_pObj = 0;
	m_color = true;
}

SxTreeNode::SxTreeNode(const SxFindable& obj, SxTreeNode *pLeft,
					   SxTreeNode *pRight) // throws SXFC_Exception
{
	m_pObj = 0;
	m_color = true;

	Allocator pfnAllocate = obj.getAllocator();
	pfnAllocate((SxObject**)(&m_pObj));
	m_pObj->assign(obj);

	m_pLeft = pLeft;
	m_pRight = pRight;
}

// SxObject
SxObject& SxTreeNode::assign(const SxObject& obj) // throws SXFC_Exception
{
	SxTreeNode *pNode = (SxTreeNode*)&obj;

	m_pLeft = pNode->m_pLeft;
	m_pRight = pNode->m_pRight;

	if(m_pObj)
	{
		delete m_pObj;
		m_pObj = 0;
	}

	if(&(pNode->obj()))
	{
		Allocator pfnAllocate = pNode->obj().getAllocator();
		pfnAllocate((SxObject**)&m_pObj);
		m_pObj->assign(pNode->obj());
	}

	return *this;
}

SxObject& SxTreeNode::operator =(const SxObject& obj)
{
	if(!!m_pObj)
		*m_pObj = obj;
	
	else
	{
		Allocator pfnAllocate = obj.getAllocator();
		pfnAllocate((SxObject**)&m_pObj);
		m_pObj->assign(obj);
	}

	return *m_pObj;
}


// SxFindable
bool SxTreeNode::operator <(const SxFindable& obj) const
{
	if(!m_pObj)
		return false;

	return (*m_pObj < ((const SxTreeNode&)obj).obj());
}

bool SxTreeNode::operator ==(const SxFindable& obj) const
{
	if(!m_pObj)
		return false;

	return (*m_pObj == obj);
}


////////////////// implementation of class SxRBTree


const int LEFT = 0;
const int RIGHT = 1;

// public
SxFindable *SxRBTree::put(const SxFindable& obj) // throws SXFC_Exception
{
	SxTreeNode *pNewNode = new SxTreeNode(obj, m_pDummy, m_pDummy);

	if(!m_pRoot)
	{
		m_pRoot = pNewNode;
		m_pRoot->m_color = false;
		m_size++;

		return &pNewNode->obj();
	}

	SxStack searchPath(ROUND_LOG(m_size) + 1);
	SxTreeNode *pNode = m_pRoot, *pParent = 0;
	bool isSmaller = false;

	while(!pNode->isDummy())
	{
		pParent = pNode;

		searchPath.push(pNode);

		if(pNode->obj() == obj)
		{
			delete pNewNode;
			throw SXFC_Exception(E_SXFC_DUPLICATE_KEYS);
			return 0;
		}
		// two comparisons; should also be improved (see class SxFindable)
		isSmaller = obj < pNode->obj() ? true : false;

		pNode = isSmaller ? pNode->m_pLeft : pNode->m_pRight;
	}

	if(!!pParent)
	{
		if(isSmaller)
		{
			pParent->m_pLeft = pNewNode;
		}
		else
		{
			pParent->m_pRight = pNewNode;
		}
	}

	m_pSecondRed = pNewNode;
	restoreBalance(searchPath);
	m_size++;
	return &pNewNode->obj();
}

const SxFindable* SxRBTree::find(const SxFindable& obj) const
{
	if(!m_size)
		return 0;

	SxTreeNode *pNode = m_pRoot;

	while(!pNode->isDummy())
	{
		if(*pNode == obj)
			return &pNode->obj();

		pNode = pNode->obj() < obj ? pNode->m_pRight : pNode->m_pLeft;
	}

	return 0;
}

// private/protected
void SxRBTree::restoreBalance(SxStack& path)
{
	int cRed = 1;
	while(path.size())
	{
		SxTreeNode *pNode = (SxTreeNode*)path.pop();
		SxTreeNode *pNewRoot = 0;
		if(pNode->m_color)
		{
			cRed++;
			if(cRed == 1)
			{
				m_pSecondRed = pNode;
			}
		}
		else
			cRed = 0;

		if(cRed == 2)
		{
			SxTreeNode *pParent = (SxTreeNode*)path.pop();
			if(changeColor(pParent))
			{
				cRed = 1;
				m_pSecondRed = pParent;
			}
			else
			{
				cRed = 0;
				bool dRotation = false;
				bool bParentSmaller = *pParent < *pNode;

				if((bParentSmaller ? *pParent : *pNode) < *m_pSecondRed
					&& *m_pSecondRed < (bParentSmaller ? *pNode : *pParent))
				{
					dRotation = true;
				}

				if(*pParent < *pNode)
				{
					pNewRoot = rotateLeft(pParent, dRotation, true);
				}
				else
				{
					pNewRoot = rotateRight(pParent, dRotation, true);
				}

				// insert the tree
				if(!path.size())
				{
					m_pRoot = pNewRoot;
				}
				else
				{
					SxTreeNode *pTop = (SxTreeNode*)path.peek();
					if(pTop->m_pLeft == pParent)
					{
						pTop->m_pLeft = pNewRoot;
					}
					else
					{
						pTop->m_pRight = pNewRoot;
					}
				}
			}
		}
		m_pRoot->m_color = false;
	}
}

SxTreeNode *SxRBTree::rotateLeft(SxTreeNode *pParent, bool dRotation, bool bWithColor)
{
	SxTreeNode *pOldRoot = pParent;
	SxTreeNode *pNewRoot;

	if(dRotation)
	{
		SxTreeNode *pNewRight;

		pNewRight = pOldRoot->m_pRight;
		pNewRoot = pNewRight->m_pLeft;
		pOldRoot->m_pRight = pNewRoot->m_pLeft;
		pNewRight->m_pLeft = pNewRoot->m_pRight;
		pNewRoot->m_pRight = pNewRight;
		pNewRoot->m_pLeft = pOldRoot;
	}
	else
	{
		pNewRoot = pOldRoot->m_pRight;
		pOldRoot->m_pRight = pNewRoot->m_pLeft;
		pNewRoot->m_pLeft = pOldRoot;
	}

	if(bWithColor)
	{
		pNewRoot->m_color = false;
		pNewRoot->m_pLeft->m_color = true;
	}

	return pNewRoot;
}

SxTreeNode *SxRBTree::rotateRight(SxTreeNode *pParent, bool dRotation, bool bWithColor)
{
	SxTreeNode *pOldRoot = pParent;
	SxTreeNode *pNewRoot;

	if(dRotation)
	{
		SxTreeNode *pNewLeft;

		pNewLeft = pOldRoot->m_pLeft;
		pNewRoot = pNewLeft->m_pRight;
		pOldRoot->m_pLeft = pNewRoot->m_pRight;
		pNewLeft->m_pRight = pNewRoot->m_pLeft;
		pNewRoot->m_pLeft = pNewLeft;
		pNewRoot->m_pRight = pOldRoot;
	}
	else
	{
		pNewRoot = pOldRoot->m_pLeft;
		pOldRoot->m_pLeft = pNewRoot->m_pRight;
		pNewRoot->m_pRight = pOldRoot;
	}

	if(bWithColor)
	{
		pNewRoot->m_color = false;
		pNewRoot->m_pRight->m_color = true;
	}

	return pNewRoot;
}


bool SxRBTree::changeColor(SxTreeNode *pNode)
{
	if(pNode->m_pLeft == 0 || pNode->m_pRight == 0)
	{
		return false;
	}

	if(!pNode->m_color && pNode->m_pLeft->m_color && pNode->m_pRight->m_color)
	{
		pNode->m_color = true;
		pNode->m_pLeft->m_color = false;
		pNode->m_pRight->m_color = false;

		return true;
	}

	return false;
}

SxStack* SxRBTree::pathTo(const SxFindable& obj) const
{
	SxStack *pSearchPath = new SxStack(ROUND_LOG(m_size) + 1);

	if(!m_size)
	{
		delete pSearchPath;
		return 0;
	}

	SxTreeNode *pNode = m_pRoot;

	while(!pNode->isDummy())
	{
		pSearchPath->push(pNode);

		if(obj == pNode->obj())
		{
			return pSearchPath;
		}
		if(obj < pNode->obj())
			pNode = pNode->m_pLeft;
		else
			pNode = pNode->m_pRight;
	}

	delete pSearchPath;
	return 0;
}

bool SxRBTree::remove(const SxFindable& obj)
{
	SxTreeNode *pX = 0, *pY = 0, *pYParent = 0;
	SxStack *pSearchPath = pathTo(obj);

	if(!pSearchPath)
		return false;

	SxTreeNode *pZ = (SxTreeNode*)pSearchPath->peek();

	if(pZ->m_pLeft->isDummy() || pZ->m_pRight->isDummy())
		pY = pZ;
	else
	{
		pZ = (SxTreeNode*)(successorOf(*pSearchPath)->peek());
	}

	pSearchPath->pop();
	if(pSearchPath->size())
	{
		pYParent = (SxTreeNode*)pSearchPath->peek();
	}
	else
		pYParent = 0;

	if(!pY->m_pLeft->isDummy())
		pX = pY->m_pLeft;
	else
		pX = pY->m_pRight;

	if(pYParent == 0)
	{
		delete m_pRoot;
		m_pRoot = pX;
	}
	else
	{
		if(pY == pYParent->m_pLeft)
		{
			delete pYParent->m_pLeft;
			pYParent->m_pLeft = pX;
		}
		else
		{
			delete pYParent->m_pRight;
			pYParent->m_pRight = pX;
		}
	}
	if(pY != pZ)
	{
		*pZ = pY->obj();
	}

	pSearchPath->push(pX);
	if(!pY->m_color)
		fixup(*pSearchPath);

	m_size--;
	delete pSearchPath;
	return true;
}

SxStack *SxRBTree::successorOf(SxStack& path) const
{
	SxTreeNode *pY = ((SxTreeNode*)path.peek())->m_pRight;
	path.push(pY);

	while(!pY->m_pLeft->isDummy())
	{
		pY = pY->m_pLeft;
		path.push(pY);
	}

	return &path;
}

void SxRBTree::fixup(SxStack& xPath)
{
	SxTreeNode *pX = 0, *pXParent = 0, *pW = 0;

	if(xPath.size() < 2)
		return;

	pX = (SxTreeNode*)xPath.pop();
	pXParent = (SxTreeNode*)xPath.pop();

	while(pX != m_pRoot && !pX->m_color)
	{
		if(pX == pXParent->m_pLeft)
		{
			pW = pXParent->m_pRight;

			if(pW->m_color)
			{
				pW->m_color = false;
				pXParent->m_color = true;
				makeBalance(pXParent, LEFT);
				pW = pXParent->m_pRight;
			}

			if(!pW->m_pLeft->m_color && !pW->m_pRight->m_color)
			{
				pW->m_color = true;
				pX = pXParent;
				SxStack *pS = pathTo(pX->obj());
				pS->pop();
				pXParent = (SxTreeNode*)pS->peek();
				delete pS;
			}
			else
			{
				if(!pW->m_pRight->m_color)
				{
					pW->m_pLeft->m_color = false;
					pW->m_color = true;
					makeBalance(pW, RIGHT);
					pW = pXParent->m_pRight;
				}
				pW->m_color = pXParent->m_color;
				pXParent->m_color = false;
				pW->m_pRight->m_color = false;
				makeBalance(pXParent, LEFT);
				pX = m_pRoot;
			}
		}
		else
		{
			pW = pXParent->m_pLeft;

			if(pW->m_color)
			{
				pW->m_color = false;
				pXParent->m_color = true;
				makeBalance(pXParent, RIGHT);
				pW = pXParent->m_pLeft;
			}

			if(!pW->m_pRight->m_color && !pW->m_pLeft->m_color)
			{
				pW->m_color = true;
				pX = pXParent;
				SxStack *pS = pathTo(pX->obj());
				pS->pop();
				pXParent = (SxTreeNode*)pS->peek();
				delete pS;
			}
			else
			{
				if(!pW->m_pLeft->m_color)
				{
					pW->m_pRight->m_color = false;
					pW->m_color = true;
					makeBalance(pW, LEFT);
					pW = pXParent->m_pLeft;
				}

				pW->m_color = pXParent->m_color;
				pXParent->m_color = false;
				pW->m_pLeft->m_color = false;
				makeBalance(pXParent, RIGHT);
				pX = m_pRoot;
			}
		}
	}

	pX->m_color = false;
}

void SxRBTree::makeBalance(SxTreeNode *pNode, int dir)
{
	SxTreeNode *pParent = 0;

	if(pNode == m_pRoot)
	{
		if(dir == LEFT)
			m_pRoot = rotateLeft(pNode, false, false);
		else
			m_pRoot = rotateRight(pNode, false, false);
	}
	else
	{
		SxStack *pS = pathTo(pNode->obj());
		pS->pop();
		pParent = (SxTreeNode*)pS->peek();
		delete pS;

		if(pParent->m_pLeft == pNode)
		{
			if(dir == LEFT)
				pParent->m_pLeft = rotateLeft(pNode, false, false);
			else
				pParent->m_pLeft = rotateRight(pNode, false, false);
		}
		else
		{
			if(dir == LEFT)
				pParent->m_pRight = rotateLeft(pNode, false, false);
			else
				pParent->m_pRight = rotateRight(pNode, false, false);
		}
	}
}

void SxRBTree::clear(SxTreeNode *pNode)
{
	if(!!pNode)
	{
		if(!pNode->isDummy())
		{
			clear(pNode->m_pLeft);
			pNode->m_pLeft = 0;
			clear(pNode->m_pRight);
			pNode->m_pRight = 0;
			delete pNode;
			m_size--;
		}
	}
}


