// sxmap.h: some non template map definitions

#ifndef __SXMAP_H__
#define __SXMAP_H__

#include "sxsort.h"
#include "sxcontainer.h"
#include "sxstack.h"
#include "sxpair.h"
#include "sxstring.h"
#include "sxlist.h"


/////////////// definition of class SxTreeNode

class SxTreeNode : public SxFindable
{
public:
	SxTreeNode();
	SxTreeNode(const SxFindable&, SxTreeNode*, SxTreeNode*); // throws SXFC_Exception;

	virtual ~SxTreeNode()
	{
		if(!!m_pObj)
			delete m_pObj;
	}

	// SxObject
	ALLOC_IMPL(SxTreeNode);
	virtual SxObject& assign(const SxObject&); // throws SXFC_Exception; copies the whole tree node
	virtual SxObject& operator =(const SxObject&); // throws SXFC_Exception; assigns only m_pObj
	virtual bool isInstanceOf(const TCHAR *pszClass) const
	{
		if(!TCSCMP(pszClass, _T("SxTreeNode")))
			return true;

		return SxFindable::isInstanceOf(pszClass);
	}


	// SxFindable
	virtual bool operator <(const SxFindable&) const;
	virtual bool operator ==(const SxFindable&) const;

	inline SxFindable& obj() const { return *m_pObj; }
	bool isDummy()
	{
		if(m_pObj)
		{
			if(m_pObj->isInstanceOf(_T("SxString")))
			{
				if(((SxString&)*m_pObj) == SxString(_T("__dummy")))
					return true;
			}
		}
		return false;
	}

	bool m_color;  // black = false, red = true;

	SxTreeNode *m_pLeft;
	SxTreeNode *m_pRight;

protected:
	SxFindable *m_pObj;
};


/////////////// definition of class SxRBTree

class SxRBTree : public SxContainer
{
public:
	SxRBTree() : SxContainer(), m_pRoot(0), m_pSecondRed(0)
	{
		m_pDummy = new SxTreeNode(SxString(_T("__dummy")), 0, 0);
		m_pDummy->m_color = false;
	}

	virtual ~SxRBTree()
	{
		this->clear(m_pRoot);
		delete m_pDummy;
	}

	// SxObject
	ALLOC_IMPL(SxRBTree);
	virtual bool isInstanceOf(const TCHAR *pszClass) const
	{
		if(!TCSCMP(pszClass, _T("SxRBTree")))
			return true;

		return SxContainer::isInstanceOf(pszClass);
	}

	virtual SxFindable *put(const SxFindable&);
	virtual bool remove(const SxFindable&);
	virtual void clear(SxTreeNode *pNode);
	virtual const SxFindable *find(const SxFindable&) const;

protected:
	void restoreBalance(SxStack&);

	SxTreeNode *rotateLeft(SxTreeNode*, bool, bool);
	SxTreeNode *rotateRight(SxTreeNode*, bool, bool);

	bool changeColor(SxTreeNode*);

	SxStack *pathTo(const SxFindable&) const;
	SxStack *successorOf(SxStack&) const;
	void fixup(SxStack& xPath);
	void makeBalance(SxTreeNode*, int);

	SxTreeNode *m_pRoot, *m_pSecondRed, *m_pDummy;
};


/////////////// definition of class SxMapItem

class SxMapItem : public SxFindable
{
public:
	SxMapItem() : SxFindable(), m_pObj(0), m_pPrevious(0), m_pNext(0) {;}
	SxMapItem(SxFindable *pObj, SxMapItem* pPrevious = 0)
		: SxFindable(), m_pObj(pObj), m_pPrevious(pPrevious), m_pNext(0)
	{
		if(pPrevious)
			pPrevious->m_pNext = this;
	}

	virtual ~SxMapItem()
	{
		if(m_pPrevious)
			m_pPrevious->m_pNext = m_pNext;

		if(m_pNext)
			m_pNext->m_pPrevious = m_pPrevious;
	}

	// SxObject
	ALLOC_IMPL(SxMapItem);
	virtual bool isInstanceOf(const TCHAR *pszClass) const
	{
		if(!TCSCMP(pszClass, _T("SxMapItem")))
			return true;

		return SxFindable::isInstanceOf(pszClass);
	}
	virtual SxObject& assign(const SxObject& obj)
	{
		if(obj.isInstanceOf(_T("SxMapItem")))
		{
			m_pObj = ((const SxMapItem&)obj).m_pObj;
			m_pPrevious = ((const SxMapItem&)obj).m_pPrevious;
			m_pNext = ((const SxMapItem&)obj).m_pNext;

			return *this;
		}
		else
		{
			if(!m_pObj)
			{
				Allocator pfnAllocate = obj.getAllocator();
				pfnAllocate((SxObject**)&m_pObj);
			}

			return m_pObj->assign(obj);
		}
	}
	virtual SxObject& operator =(const SxFindable& obj)
	{
		return assign(obj);
	}

	// SxFindable
	virtual bool operator <(const SxFindable& obj) const
	{
		if(obj.isInstanceOf(_T("SxMapItem")))
			return (*m_pObj < *((SxMapItem&)obj).m_pObj);
		else
			return (*m_pObj < obj);
	}

	inline virtual bool operator ==(const SxFindable& obj) const {	return (*m_pObj == obj); }

	SxMapItem *m_pPrevious, *m_pNext;
	SxFindable *m_pObj;
};


/////////////// definition of class SxMap

class SxMap : public SxRBTree
{
public:
	class iterator
	{
	public:
		iterator() : m_pItem(0) {;};
		iterator(SxMapItem* pItem)
		{
			*this = pItem;
		}
		iterator& operator ++()
		{
			if(m_pItem)
				m_pItem = m_pItem->m_pNext;
			return *this;
		}
		iterator operator ++(int)
		{
			iterator tmp = *this;
			++(*this);
			return tmp;
		}
		iterator& operator --()
		{
			if(m_pItem)
				m_pItem = m_pItem->m_pPrevious;
			return *this;
		}
		iterator operator --(int)
		{
			iterator tmp = *this;
			--(*this);
			return tmp;
		}
		inline SxKeyValue* operator ->()
		{
			return (SxKeyValue*)m_pItem->m_pObj;
		}
		iterator& operator =(SxMapItem *pMapItem)
		{
			m_pItem = pMapItem;
			return *this;
		}
		inline SxKeyValue& operator *()
		{
			return *(SxKeyValue*)m_pItem->m_pObj;
		}
		inline bool operator ==(SxMapItem *pMapItem) const
		{
			return (m_pItem == pMapItem);
		}
		inline bool operator !=(SxMapItem *pMapItem) const
		{
			return (m_pItem != pMapItem);
		}
		inline bool operator ==(iterator& it) const
		{
			return (m_pItem == it.m_pItem);
		}
		inline bool operator !=(iterator& it) const
		{
			return (m_pItem != it.m_pItem);
		}

		SxMapItem *m_pItem;
	};

	SxMap() : SxRBTree(), m_pFirst(0), m_pLast(0) {;}

	virtual ~SxMap()
	{
		clear(m_pRoot);
	}

	// SxObject
	ALLOC_IMPL(SxMap);
	virtual bool isInstanceOf(const TCHAR *pszClass) const
	{
		if(!TCSCMP(pszClass, _T("SxMap")))
			return true;

		return SxRBTree::isInstanceOf(pszClass);
	}

	// SxRBTree
	virtual SxFindable *put(const SxFindable& obj) // throws SXFC_Exception
	{
		SxMapItem *pItem = new SxMapItem();

		Allocator pfnAllocate = obj.getAllocator();
		pfnAllocate((SxObject**)&pItem->m_pObj);
		pItem->m_pObj->assign(obj);
		SxMapItem *pRet = (SxMapItem*)SxRBTree::put(*pItem);
		delete pItem;

		if(!m_pLast)
		{
			m_pFirst = pRet;
			m_pLast = pRet;
		}
		else
		{
			pRet->m_pPrevious = m_pLast;
			m_pLast->m_pNext = pRet;
			m_pLast = pRet;
		}

		return pRet;
	}

	virtual void clear(SxTreeNode *pNode)
	{
		if(!!pNode)
		{
			if(!pNode->isDummy())
			{
				clear(pNode->m_pLeft);
				clear(pNode->m_pRight);
				delete ((SxMapItem*)&pNode->obj())->m_pObj;
				((SxMapItem*)&pNode->obj())->m_pObj = 0;
			}
		}
	}

	virtual SxFindable& operator[](const SxFindable& obj)
	{
		SxMapItem *pF = (SxMapItem*)find(obj);
		if(!!pF)
			return *pF->m_pObj;

		pF = (SxMapItem*)put(obj);

		return *pF->m_pObj;
	}

	virtual iterator insert(const SxFindable& obj)
	{
		SxMapItem *pItem = (SxMapItem*)put(obj);

		return iterator(pItem);
	}

	virtual SxMapItem *begin()
	{
		return m_pFirst;
	}
	virtual SxMapItem *end()
	{
		return 0;
	}

protected:
	SxMapItem *m_pFirst, *m_pLast;
};


/////////////// definition of class SxMultimap

#define SXMUTLTIMAP_INITIAL_SIZE 5L

class SxMultimap : public SxRBTree
{
public:
	class iterator
	{
	public:
		iterator() : m_pItem(0) {;};
		iterator(SxMapItem* pItem)
		{
			*this = pItem;
		}
		iterator& operator ++()
		{
			if(m_pItem)
				m_pItem = m_pItem->m_pNext;
			return *this;
		}
		iterator operator ++(int)
		{
			iterator tmp = *this;
			++(*this);
			return tmp;
		}
		iterator& operator --()
		{
			if(m_pItem)
				m_pItem = m_pItem->m_pPrevious;
			return *this;
		}
		iterator operator --(int)
		{
			iterator tmp = *this;
			--(*this);
			return tmp;
		}
		inline SxKeyValue* operator ->()
		{
			return (SxKeyValue*)m_pItem->m_pObj;
		}
		iterator& operator =(SxMapItem *pMapItem)
		{
			m_pItem = pMapItem;
			return *this;
		}
		inline SxKeyValue& operator *()
		{
			return *(SxKeyValue*)m_pItem->m_pObj;
		}
		inline bool operator ==(SxMapItem *pMapItem) const
		{
			return (m_pItem == pMapItem);
		}
		inline bool operator !=(SxMapItem *pMapItem) const
		{
			return (m_pItem != pMapItem);
		}
		inline bool operator ==(iterator& it) const
		{
			return (m_pItem == it.m_pItem);
		}
		inline bool operator !=(iterator& it) const
		{
			return (m_pItem != it.m_pItem);
		}

		SxMapItem *m_pItem;
	};

	SxMultimap() : SxRBTree(), m_pFirst(0), m_pLast(0) {;}

	virtual ~SxMultimap()
	{
		this->clear(m_pRoot);
	}

	// SxObject
	ALLOC_IMPL(SxMultimap);
	virtual bool isInstanceOf(const TCHAR *pszClass)
	{
		if(!TCSCMP(pszClass, _T("SxMultimap")))
			return true;

		return SxRBTree::isInstanceOf(pszClass);
	}

	// SxRBTree
	virtual SxFindable *put(const SxFindable& obj)
	{
		SxFindableVector *pVec = (SxFindableVector*)find(obj), *pRet;
		Allocator pfnAllocate = obj.getAllocator();

		if(!obj.isInstanceOf(_T("SxFindableVector")))
		{
			SxMapItem *pItem = new SxMapItem();

			pfnAllocate((SxObject**)&pItem->m_pObj);
			pItem->m_pObj->assign(obj);

			if(!m_pLast)
			{
				m_pFirst = pItem;
				m_pLast = pItem;
			}
			else
			{
				pItem->m_pPrevious = m_pLast;
				m_pLast->m_pNext = pItem;
				m_pLast = pItem;
			}

			if(!!pVec)
			{
				pVec->push_back(pItem);
				m_size++;
				return pItem;
			}
			pVec = new SxFindableVector(SXMUTLTIMAP_INITIAL_SIZE);
			pVec->push_back(pItem);
			SxRBTree::put(*pVec);
			m_size++;
			delete pVec;
			return pItem;
		}
		else
		{
			if(!!pVec)
			{
				throw SXFC_Exception(E_SXFC_DUPLICATE_KEYS);
				return 0;
			}
			pRet = (SxFindableVector*)SxRBTree::put(obj);
			m_size += ((SxFindableVector&)obj).size();
			return pRet;
		}
	}

	virtual void clear(SxTreeNode *pNode)
	{
		if(!!pNode)
		{
			if(!pNode->isDummy())
			{
				clear(pNode->m_pLeft);
				clear(pNode->m_pRight);
				SxFindableVector *pVec = (SxFindableVector*)&pNode->obj();
				while(pVec->size())
				{
					delete ((SxMapItem*)pVec->at(pVec->size() - 1))->m_pObj;
					delete pVec->at(pVec->size() - 1);
					pVec->pop_back();
				}
			}
		}
	}


	virtual SxFindableVector& operator[](const SxFindable& obj)
	{
		SxFindableVector *pVec = (SxFindableVector*)find(obj), *pRet;
		if(!!pVec)
			return *pVec;

		SxKeyValue *pKeyVal = new SxKeyValue(obj, SxString(_T("")));
		pVec = new SxFindableVector(SXMUTLTIMAP_INITIAL_SIZE);
		pVec->push_back(pKeyVal);
		pRet = (SxFindableVector*)put(*pVec);
		delete pVec;
		return *pRet;
	}

	virtual bool remove(const SxFindable& obj, long nIndex = 0)
	{
		SxFindableVector *pVec = (SxFindableVector*)find(obj);
		bool bRetVal = false;

		if(!pVec)
			return bRetVal;

		if(nIndex > (long)pVec->size() - 1)
			return bRetVal;

		SxKeyValue *pKeyVal = (SxKeyValue*)((SxMapItem*)pVec->at(nIndex))->m_pObj;

		if(pVec->size() == 0)
		{
			bRetVal = SxRBTree::remove(obj);
		}
		else
		{
			bRetVal = true;
			delete (SxMapItem*)pVec->at(nIndex);
			pVec->remove(nIndex);
		}

		delete pKeyVal;
		m_size--;
		return bRetVal;
	}

	virtual iterator insert(const SxFindable& obj)
	{
		SxMapItem *pItem = (SxMapItem*)put(obj);

		return iterator(pItem);
	}

	virtual SxMapItem *begin()
	{
		return m_pFirst;
	}
	virtual SxMapItem *end()
	{
		return 0;
	}

protected:
	SxMapItem *m_pFirst, *m_pLast;
};


/////////////// definition of class SxStrMap
class SxStrMap : public SxMap
{
public:
	SxStrMap() : SxMap()
	{
		;
	}

	// SxObject
	ALLOC_IMPL(SxStrMap);
	virtual bool isInstanceOf(const TCHAR *pszClass) const
	{
		if(!TCSCMP(pszClass, _T("SxStrMap")))
			return true;

		return SxMap::isInstanceOf(pszClass);
	}

	virtual SxString& operator[](const SxString& obj)
	{
		SxMapItem *pItem = (SxMapItem*)find(obj);
		SxKeyValue *pKeyVal = 0;

		if(pItem)
			pKeyVal = (SxKeyValue*)pItem->m_pObj;

		if(!!pKeyVal)
			return (SxString&)pKeyVal->second();

		pKeyVal = new SxKeyValue(obj, SxString(_T("")));
		SxMapItem *pRet = (SxMapItem*)put(*pKeyVal);
		delete pKeyVal;

		return (SxString&)((SxKeyValue*)pRet->m_pObj)->second();
	}
};


/////////////// definition of class SxStrObjMap

class SxStrObjMap : public SxMap
{
public:
	SxStrObjMap() : SxMap()
	{
		;
	}

	// SxObject
	ALLOC_IMPL(SxStrObjMap);
	virtual bool isInstanceOf(const TCHAR *pszClass) const
	{
		if(!TCSCMP(pszClass, _T("SxStrObjMap")))
			return true;

		return SxMap::isInstanceOf(pszClass);
	}

	virtual SxObject& operator[](const SxString& obj)
	{
		SxKeyValue *pKeyVal = (SxKeyValue*)find(obj);
		if(!!pKeyVal)
			return pKeyVal->second();

		return *put(SxKeyValue(obj, SxObject()));
	}
};


/////////////// definition of class SxStrObjMultimap

class SxStrObjMultimap : public SxMultimap
{
public:
	SxStrObjMultimap() : SxMultimap()
	{
		;
	}

	// SxObject
	ALLOC_IMPL(SxStrObjMultimap);
	virtual bool isInstanceOf(const TCHAR *pszClass)
	{
		if(!TCSCMP(pszClass, _T("SxStrObjMultimap")))
			return true;

		return SxMultimap::isInstanceOf(pszClass);
	}

	virtual SxFindable *put(const SxString& strKey, const SxObject& value)
	{
		return SxMultimap::put(SxKeyValue(strKey, value));
	}
};



/////////////// definition of class SxStrPtrMultimap

class SxStrPtrMultimap : public SxMultimap
{
public:
	class iterator
	{
	public:
		iterator() : m_pItem(0) {;};
		iterator(SxMapItem* pItem)
		{
			*this = pItem;
		}
		iterator& operator ++()
		{
			if(m_pItem)
				m_pItem = m_pItem->m_pNext;
			return *this;
		}
		iterator operator ++(int)
		{
			iterator tmp = *this;
			++(*this);
			return tmp;
		}
		iterator& operator --()
		{
			if(m_pItem)
				m_pItem = m_pItem->m_pPrevious;
			return *this;
		}
		iterator operator --(int)
		{
			iterator tmp = *this;
			--(*this);
			return tmp;
		}
		inline SxStrPtr* operator ->()
		{
			return (SxStrPtr*)m_pItem->m_pObj;
		}
		iterator& operator =(SxMapItem *pMapItem)
		{
			m_pItem = pMapItem;
			return *this;
		}
		inline SxStrPtr& operator *()
		{
			return *(SxStrPtr*)m_pItem->m_pObj;
		}
		inline bool operator ==(SxMapItem *pMapItem) const
		{
			return (m_pItem == pMapItem);
		}
		inline bool operator !=(SxMapItem *pMapItem) const
		{
			return (m_pItem != pMapItem);
		}
		inline bool operator ==(iterator& it) const
		{
			return (m_pItem == it.m_pItem);
		}
		inline bool operator !=(iterator& it) const
		{
			return (m_pItem != it.m_pItem);
		}

		SxMapItem *m_pItem;
	};

	SxStrPtrMultimap() : SxMultimap()
	{
		;
	}

	// SxObject
	ALLOC_IMPL(SxStrPtrMultimap);
	virtual bool isInstanceOf(const TCHAR *pszClass)
	{
		if(!TCSCMP(pszClass, _T("SxStrPtrMultimap")))
			return true;

		return SxMultimap::isInstanceOf(pszClass);
	}

	virtual SxFindable *put(const SxString& strKey, const SxPtr& p)
	{
		return SxMultimap::put(SxStrPtr(strKey, p));
	}

	virtual iterator insert(const SxStrPtr& p)
	{
		SxMapItem *pItem = (SxMapItem*)put(p.first(), p.second());

		return iterator(pItem);
	}

};




#endif // __SXMAP_H__
