// sxstring.h: sxfc string class definition

#ifndef __SXSTRING_H__
#define __SXSTRING_H__

#include "sxsort.h"
#include "sxexcept.h"
#include "sxfcerr.h"
#include "sxtchar.h"

// minimum capacity, which an SxString grows
#define SXSTRING_RESERVE_CAPACITY 64L


class SxString : public SxFindable
{
public:
	typedef SX_SIZE_T size_type;

	class const_iterator
	{
	public:
		const_iterator& operator ++()
		{
			m_p++;
			return *this;
		}
		const_iterator operator ++(int)
		{
			const_iterator tmp = *this;
			++(*this);
			return tmp;
		}
		const_iterator& operator --()
		{
			m_p--;
			return *this;
		}
		const_iterator operator --(int)
		{
			const_iterator tmp = *this;
			--(*this);
			return tmp;
		}
		inline bool operator ==(const const_iterator& it) const
		{
			return (m_p == (const TCHAR*)(it));
		}
		inline bool operator !=(const const_iterator& it) const
		{
			return (m_p != (const TCHAR*)(it));
		}
		const_iterator& operator =(const const_iterator& it)
		{
			m_p = (const TCHAR*)(it);
			return *this;
		}
		const_iterator& operator =(const TCHAR *psz)
		{
			m_p = psz;
			return *this;
		}
		inline const TCHAR& operator *() const
		{
			return *m_p;
		}
		inline operator const TCHAR *() const { return m_p; }

	protected:
		const TCHAR *m_p;
	};

	class iterator
	{
	public:
		iterator& operator ++()
		{
			m_p++;
			return *this;
		}
		iterator operator ++(int)
		{
			iterator tmp = *this;
			++(*this);
			return tmp;
		}
		iterator& operator --()
		{
			m_p--;
			return *this;
		}
		iterator operator --(int)
		{
			iterator tmp = *this;
			--(*this);
			return tmp;
		}
		inline bool operator ==(const iterator& it) const
		{
			return (m_p == (TCHAR*)(it));
		}
		inline bool operator !=(const iterator& it) const
		{
			return (m_p != (TCHAR*)(it));
		}
		iterator& operator =(const iterator& it)
		{
			m_p = (TCHAR*)(it);
			return *this;
		}
		iterator& operator =(TCHAR *psz)
		{
			m_p = psz;
			return *this;
		}
		inline TCHAR& operator *()
		{
			return *m_p;
		}
		inline operator TCHAR*() const { return m_p; }

	protected:
		TCHAR *m_p;
	};

	// construction/destruction
	SxString() : m_p(0) // throws SXFC_Exception
	{
		this->clear();
	}

	SxString(const TCHAR *psz)
		: m_p(0) // throws SXFC_Exception
	{
		this->clear();
		this->assign(psz);
	}

	SxString(const TCHAR *psz, size_type len)
		: m_p(0) // throws SXFC_Exception
	{
		this->clear();
		this->assign(psz, len);
	}

	SxString(const SxString& str)
		: m_p(0) // throws SXFC_Exception
	{
		this->clear();
		this->assign(str);
	}

	SxString(const SxString& base, size_type offset, size_type len)
		: m_p(0) // throws SXFC_Exception
	{
		this->clear();
		this->assign(base.substr(offset, len));
	}

	virtual ~SxString()
	{
		this->clear();
	}


	// SxObject
	ALLOC_IMPL(SxString);
	virtual SxObject& assign(const SxObject&); // throws SXFC_Exception
	virtual bool isInstanceOf(const TCHAR *pszClass) const
	{
		if(!TCSCMP(pszClass, _T("SxString")))
			return true;

		return SxFindable::isInstanceOf(pszClass);
	}

	// SxFindable
	inline virtual bool operator <(const SxFindable& str) const
	{
		return compare((const SxString&)str) < 0 ? true : false;
	}

	inline virtual bool operator ==(const SxFindable& str) const
	{
		return !compare((const SxString&)str) ? true : false;
	}

	// operators
	inline virtual bool operator ==(const TCHAR *psz) const
	{
		return !compare(psz) ? true : false;
	}

	inline bool operator !=(const SxString& str) const
	{
		return compare(str) ? true : false;
	}

	inline bool operator !=(const TCHAR *psz) const
	{
		return compare(psz) ? true : false;
	}


	inline SxString& operator =(const SxString& str) // throws SXFC_Exception
	{
		return (SxString&)this->assign((const SxObject&)(str));
	}

	inline SxString& operator +=(const SxString& str) // throws SXFC_Exception
	{
		return this->append(str);
	}

	inline SxString& operator +=(const TCHAR *psz) // throws SXFC_Exception
	{
		return this->append(psz);
	}

	inline SxString& operator +=(const TCHAR c) // throws SXFC_Exception
	{
		return this->append(c);
	}

	inline TCHAR& operator[](size_type pos) // throws SXFC_Exception
	{
		if(pos >= m_size)
			throw SXFC_Exception(E_SXFC_OUT_OF_BOUND);

		return m_p[pos];
	}

//	inline operator const TCHAR*() const { return m_p; }

	// public methods
	SxString& append(const SxString&); // throws SXFC_Exception
	SxString& append(const TCHAR*); // throws SXFC_Exception

	SxString& assign(const TCHAR*); // throws SXFC_Exception
	SxString& assign(const TCHAR*, size_type len); // throws SXFC_Exception

	int compare(const SxString&) const;
	int compare(const TCHAR *psz) const;

	size_type find_first_of(const TCHAR *psz, size_type pos = 0) const
	{
		const TCHAR *p = m_p;
		long len = TCSLEN(psz);
		while(pos + len <= m_size)
		{
			if(!TCSNCMP(&m_p[pos], psz, len))
				return pos;

			pos++;
		}

		return npos;
	}

	SxString substr(size_type pos = 0, size_type len = npos) const
	{
		SxString ret(_T(""));

		if(pos < m_size)
		{
			if(len == npos)
				ret = &m_p[pos];

			else
			{
				while(pos < m_size && len--)
				{
					ret += m_p[pos++];
				}
			}
		}

		return ret;
	}

	// data accessor methods
	inline size_type size() const { return m_size; }
	inline size_type capacity() const { return m_capacity; }
	inline const TCHAR *c_str() const { return m_p; }
	inline bool empty() const { return m_size ? false : true; }
	inline TCHAR at(size_type pos) const
	{
		if(pos >= m_size)
			throw SXFC_Exception(E_SXFC_OUT_OF_BOUND);

		return m_p[pos];
	}

	const_iterator begin() const
	{
		const_iterator it;
		it = m_p;
		return it;
	}
	iterator begin()
	{
		iterator it;
		it = m_p;
		return it;
	}
	const_iterator end() const
	{
		const_iterator it;
		it = m_p + m_size;
		return it;
	}
	iterator end()
	{
		iterator it;
		it = m_p + m_size;
		return it;
	}

	static const size_type npos;

private:
	SxString& append(const TCHAR c);

	void clear();
	void grow(size_type); // throws SXFC_Exception

	TCHAR *m_p;
	size_type m_size;
	size_type m_capacity;
};

SxString operator +(const SxString& lhs, const SxString& rhs);
SxString operator +(const TCHAR*, const SxString&);


#endif // __SXSTRING_H__
