// xmlElement.cpp


#include "xmlcore.h"
#include "xmlElement.h"
#include "xmlElementEnum.h"


// xml Encoding
void XMLEncode(const SxString& strIn, SxString& strOut)
{
	SxString::const_iterator iIn;
	strOut = _T("");

	for(iIn = strIn.begin(); iIn != strIn.end(); ++iIn)
	{
		switch(*iIn)
		{
		case _T('&'):
			strOut += _T("&amp;");
			break;
		case _T('<'):
			strOut += _T("&lt;");
			break;
		case _T('>'):
			strOut += _T("&gt;");
			break;
		case _T('"'):
			strOut += _T("&quot;");
			break;
		case _T('\''):
			strOut += _T("&apos;");
			break;
		default:
			strOut += *iIn;
		}
	}
}

void XMLDecode(const SxString& strIn, SxString& strOut)
{
	SxString::const_iterator iIn;
	strOut = _T("");
	for(iIn = strIn.begin(); iIn != strIn.end(); ++iIn)
	{
		if(*iIn == _T('&'))
		{
			int i;
			TCHAR esc[6];
			for(i = 0; (i < 5) && (*iIn != _T(';')); ++i)
				esc[i] = *iIn++;

			esc[i] = _T('\0');

			if(TCSCMP(esc, _T("&amp")) == 0)
			{
				strOut += _T('&');
			}
			else if(TCSCMP(esc, _T("&lt")) == 0)
			{
				strOut += _T('<');
			}
			else if(TCSCMP(esc, _T("&gt")) == 0)
			{
				strOut += _T('>');
			}
			else if(TCSCMP(esc, _T("&quot")) == 0)
			{
				strOut += _T('"');
			}
			else if(TCSCMP(esc, _T("&apos")) == 0)
			{
				strOut += _T('\'');
			}
		}
		else
			strOut += *iIn;
	}
}





bool xmlElement::setAttribute(const SxString & strName, const SxString & strValue)
{
	SxString strEncoded;
	XMLEncode(strValue, strEncoded);
	m_attributes[strName] = strEncoded;
	return true;
}

bool xmlElement::getAttribute(const SxString & strName, SxString & strValue)
{
	SxMapItem *pItem = (SxMapItem*)m_attributes.find(strName);
	SxString *pValue = 0;
	if(pItem)
	{
		SxKeyValue *pKeyVal = (SxKeyValue*)pItem->m_pObj;
		if(pKeyVal)
			pValue = (SxString*)&pKeyVal->second();
	}

	if(pValue)
	{
		SxString strEncoded = *pValue;
		XMLDecode(strEncoded, strValue);
		return true;
	}
	else
	{
		return false;
	}
}

void xmlElement::insertChild(xmlElement * pElement)
{
	_ASSERT(pElement);
	pElement->m_parent = this;

	m_children.insert(SxStrPtr(pElement->m_tag, element_ptr(pElement)));
}

bool xmlElement::removeChild(const SxString & strTag, long nIndex)
{
	return m_children.remove(strTag, nIndex);
}

bool xmlElement::getChild(const SxString & strTag, long nIndex, 
						  xmlElement ** ppElement)
{
	try
	{
		const SxFindableVector * pFindVector = (SxFindableVector *) m_children.find(strTag);
		if(!pFindVector) return false;

		SxFindable * pFind = pFindVector->at(nIndex);
		SxMapItem * pItem = (SxMapItem*)pFind;
		SxStrPtr * pStrPtr = (SxStrPtr*)(pItem->m_pObj);
		SxStrPtr *pKeyVal = pStrPtr;


		//SxStrPtr *pKeyVal = (SxStrPtr*)((SxMapItem*)m_children[strTag].at(nIndex))->m_pObj;
		*ppElement = (xmlElement*)pKeyVal->second().ptr();
		(*ppElement)->AddRef();

		return true;
	}
	catch(SXFC_Exception&)
	{
		return false;
	}
}

bool xmlElement::getParent(xmlElement ** ppElement)
{
	if(!m_parent) return false;

	element_ptr pParent = m_parent;
	*ppElement = (xmlElement*)pParent.Detach();
	return true;
}

void xmlElement::getText(SxString & strText) const
{
	strText = m_strText;
}

void xmlElement::setText(const SxString & strText)
{
	m_strText = strText;
}

bool xmlElement::getTag(SxString & strTag)
{
	strTag = m_tag;
	return true;
}

bool xmlElement::toStream(OSTREAM & stmOut)
{
	bool hasTag = !(m_tag.empty());

	if(hasTag) 	
	{ 	
		stmOut << _T("<") << m_tag;

		// export attributes
		attribute_map::iterator i = m_attributes.begin();
		while(i != m_attributes.end())
		{
			stmOut << _T(" ") << (SxString&)i->first();
			if(!((SxString&)i->second()).empty())
			{
				stmOut << _T("=\"") << (SxString&)i->second() << _T("\"");
			}
			else
			{
				stmOut << _T("=\"\"");
			}
				
			++i;
		}
	}

	bool hasBody = false;
	if(m_strText != _T(""))
	{
		if(hasTag) stmOut << _T(">");
		stmOut << m_strText;
		hasBody = true;
	}
	
	// export subitems
	element_map::iterator j = m_children.begin();
	if(j != m_children.end())
	{
		if(!hasBody)
		{
			if(hasTag) stmOut << _T(">");
			hasBody = true;
		}
		while(j != m_children.end())
		{
			((element_ptr&)j->second())->toStream(stmOut);
			++j;
		}
	}


	if(hasTag && hasBody) 
	{
		stmOut << _T("</") << m_tag << _T(">");
	}
	else 
	{
		if(hasTag) stmOut << _T("/>");
	}

	return true;
}

void xmlElement::enumChildren(const SxString & strTag, 
							  xmlElementEnum ** ppEnum)
{
	xmlEnumPtr pEnum = new xmlElementEnum;


	// internal creation
	pEnum->create(this, strTag);

	// copy result
	*ppEnum = pEnum.Detach();
}

bool xmlAtoi(const SxString & s, int & n)
{
	SxString::const_iterator i;
	n = 0;

	for(i = s.begin(); 
		i != s.end(); 
		++i)
	{
		if(('0' <= * i) && (* i <= '9'))
		{
			n *= 10; n += (* i) - '0';
		}
		else
			return false;
	}
	return true;
}

int xmlQuerySubElement(xmlElement * pElement, SxString & strId, 
						xmlElement ** ppElement)
{
	int begin = 0, end;

	// stores the pointer while decending the tree
	xmlElement::element_ptr current = pElement;

	SxString id = strId;
	while((end = id.find_first_of(_T("(."), begin)) 
		!= -1)
	{
		SxString tag = id.substr(begin, end - begin);
		int index = 0;

		if(id[end] == _T('('))
		{
			int t;

			// find closing bracket
			if((t = id.find_first_of(_T(")"), end + 1)) 
				== -1)
				return XMLE_INVALID_STRUCTURE;

			// get the index
			SxString s = id.substr(end + 1, t - end - 1);
			if(!xmlAtoi(s, index))
				return XMLE_INVALID_CHAR;

			end = t + 1;
		}

		// lookup child
		xmlElement::element_ptr tmp;
		if(!current->getChild(tag, index, tmp.Ptr()))
			return XMLE_TAG_NOT_FOUND;

		current = tmp;

		// skip the dot
		begin = end + 1;
	}

	// return the rest
	strId = id.substr(begin);

	// detach result
	* ppElement = current.Detach();

	return 0;
}

bool xmlGetAttributeEx(xmlElement * pElement, const SxString & strId, 
					   SxString & strValue)
{
	xmlElement::element_ptr pSub;
	SxString id(strId);
	if(xmlQuerySubElement(pElement, id, pSub.Ptr()))
		return false;

	return pSub->getAttribute(id, strValue);
}

bool xmlSetAttributeEx(xmlElement * pElement, const SxString & strId, 
					   const SxString & strValue)
{
	xmlElement::element_ptr pSub;
	SxString id(strId);
	if(xmlQuerySubElement(pElement, id, pSub.Ptr()))
		return false;

	return pSub->setAttribute(id, strValue);
}



